"""Main Restic objects and functions"""

from typing import Literal, TextIO, Union, Optional, overload
from collections.abc import Generator
from pathlib import PurePath
import json
from urllib.parse import quote as url_quote
import functools
from subprocess import CalledProcessError, PIPE
import logging
import boto3
from botocore.client import Config as BotoConfig
from botocore.exceptions import BotoCoreError, ClientError
from boto3.resources.base import ServiceResource as S3Bucket
import cproc

from .data import (
    ResticRepo,
    Snapshot,
    ResticCmd,
    SnapPath,
    SnapFile,
    SnapDir,
    Backup,
    SQLBackupGroup,
    SQLBackupItem,
    UserBackupDicts,
    UserBackups,
    _tag_quote,
)
from .exc import (
    ResticError,
    ResticBadIndexError,
    ResticLockedError,
)

RESTIC_PATH = '/usr/bin/restic'


def locked_retry(func):
    """Wraps a function and makes it automatically retry on ResticLockedError"""

    @functools.wraps(func)
    def _locked_retry(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except ResticLockedError as exc:
            if not exc.unlock_ok():
                raise
        self.unlock()
        return func(self, *args, **kwargs)

    return _locked_retry


class Restic:
    """Handles restic commands

    Args:
        endpoint (str): S3 server base url with https:// (excluding s3: prefix)
        cluster (str | None): Optional cluster name. Only used by repr()
        repo (ResticRepo): dataclass containing bucket name and key info
        lim (int, float, ProcLimit, None): max load limit. This can be a static
            limit if you set it to a float or int, or for more flexibility,
            use a ``ProcLimit`` instance. If unset, processes will not be
            ratelimited
        tmp_dir (Path | str): TMPDIR to use. Defaults to '/var/tmp/restic'.
        cache_dir (Path, str | None): RESTIC_CACHE_DIR to use. Set this to
            None to disable cache. Defaults to '/var/cache/restic/%(bucket)s'
        gomaxprocs (int, None): max CPU cores to allow restic to use
    """

    __module__ = 'restic'

    def __init__(
        self,
        *,
        endpoint: str,
        cluster: Union[str, None] = None,
        repo: ResticRepo,
        lim: Union[int, float, cproc.ProcLimit, None] = None,
        tmp_dir: Union[PurePath, str] = '/var/tmp/restic',
        cache_dir: Union[PurePath, str, None] = '/var/cache/restic/%(bucket)s',
        gomaxprocs: Union[int, None] = 2,
    ):
        self.endpoint = endpoint
        self.cluster = cluster
        self.repo = repo
        self.lim = lim
        self.tmp_dir = str(tmp_dir)
        self.gomaxprocs = gomaxprocs
        if cache_dir is None:
            self.cache_dir = None
        else:
            self.cache_dir = str(cache_dir) % {'bucket': repo.bucket}
        self.rebuild_index = self.repair_index

    def __repr__(self):
        if self.cluster:
            return f'Restic<{self.cluster}:{self.repo.bucket}>'
        return f'Restic<{self.repo.bucket}>'

    @property
    def env(self) -> dict[str, str]:
        """Restic environment variable dict"""
        ret = {
            'TMPDIR': self.tmp_dir,
            'GOGC': '1',
            'RESTIC_PASSWORD': self.repo.restic_pass,
            'AWS_ACCESS_KEY_ID': self.repo.access_key,
            'AWS_SECRET_ACCESS_KEY': self.repo.secret_key,
        }
        if self.cache_dir:
            ret['RESTIC_CACHE_DIR'] = self.cache_dir
        if self.gomaxprocs is not None and self.gomaxprocs > 0:
            ret['GOMAXPROCS'] = str(self.gomaxprocs)
        return ret

    def s3_bucket(self, timeout=30, retries=2) -> S3Bucket:
        """Gets a boto3 s3.Bucket for this repo

        Args:
            timeout (int): network timeout in secs
            retries (int): times to retry on timeout

        Returns:
            s3.Bucket: boto3 resource for this bucket
        """
        config = BotoConfig(
            connect_timeout=timeout,
            retries={'max_attempts': retries},
            read_timeout=timeout,
        )
        return boto3.resource(
            's3',
            endpoint_url=self.endpoint,
            aws_access_key_id=self.repo.access_key,
            aws_secret_access_key=self.repo.secret_key,
            config=config,
        ).Bucket(self.repo.bucket)

    def init_key_exists(self) -> Union[bool, None]:
        """Check if the /keys/ path has items in it, meaning that restic
        has been initialized

        Raises:
            botocore.exceptions.ClientError: connection problem
            botocore.exceptions.BotoCoreError: boto3 error

        Returns:
            (bool | None): whether the keys path existed in the bucket, or
            None if the bucket did not exist at all
        """
        try:
            return (
                len(list(self.s3_bucket().objects.filter(Prefix='keys/'))) > 0
            )
        except ClientError as exc:
            if 'NoSuchBucket' in str(exc):
                return None
            raise

    def prepare(self) -> bool:
        """Prepare access to this repo if needed and preemptively fix
        issues connecting to it. Plugins should probably run this first

        Returns:
            bool: whether this repo is ready for access
        """
        try:
            self.init_repo()
        except (BotoCoreError, ClientError) as exc:
            logging.error('%s %s: %s', self, type(exc).__name__, exc)
            return False
        except ResticLockedError as exc:
            if not exc.unlock_ok():
                logging.error(
                    '%s: could not automatically unlock\n%s', exc.name, exc
                )
                return False
            try:
                self.unlock()
            except ResticError:
                pass  # we'll try init_repo() one more time
        except ResticBadIndexError as exc:
            try:
                self.rebuild_index()
            except ResticError:
                pass  # we'll try init_repo() one more time
        except ResticError as exc:
            logging.error('%s: %s', exc.name, exc)
            return False
        try:
            self.init_repo()
        except (BotoCoreError, ClientError) as exc:
            logging.error('%s %s: %s', self, type(exc).__name__, exc)
            return False
        except ResticError as exc:
            logging.error('%s: %s', exc.name, exc)
            return False
        return True

    def init_repo(self) -> bool:
        """Initializes a restic repo if it hasn't been done already

        Raises:
            botocore.exceptions.ClientError: connection problem
            botocore.exceptions.BotoCoreError: boto3 error
            ResticError: if the restic init command fails with an error

        Returns:
            bool: True if initialized or False if already done
        """
        if exists := self.init_key_exists():
            return False
        if exists is None:  # bucket was missing
            bucket = self.s3_bucket()
            bucket.create()
            bucket.wait_until_exists()
        ret = self.build('init').run(stdout=PIPE, no_lim=True)
        if 'already initialized' in ret.stderr:
            return False
        if 'created restic repository' not in ret.stdout:
            raise ResticError(ret)
        return True

    def unlock(self):
        """Run restic unlock

        Raises:
            ResticError: if the restic unlock command fails
        """
        try:
            self.build('unlock').run(check=True, no_lim=True)
        except CalledProcessError as exc:
            raise ResticError(exc) from exc

    @locked_retry
    def repair_index(self):
        """Run restic repair index

        Raises:
            ResticError: if the command fails
        """
        # There's also a 'repair packs' and 'repair snapshots' command, but
        # automating that is tricky and dangerous, so we're leaving that to
        # manual troubleshooting only.
        try:
            self.build('repair', 'index', '--read-all-packs').run(
                check=True, no_lim=True
            )
        except CalledProcessError as exc:
            raise ResticError(exc) from exc

    @locked_retry
    def prune(self, no_lim: bool = True):
        """Run restic prune

        Args:
            no_lim (bool): do not CPU limit the command as it runs regardless
                of the lim arg in Restic

        Raises:
            ResticError: if the command fails
        """
        if self.cache_dir:
            args = ('prune', '--cleanup-cache')
        else:
            args = ('prune',)
        try:
            self.build(*args).run(check=True, no_lim=no_lim)
        except CalledProcessError as exc:
            raise ResticError(exc) from exc

    def build(self, *args) -> ResticCmd:
        """Build a ResticCmd object that can be used to execute the requested
        restic command

        Args:
            *args (tuple[str]): restic subcommand and args

        Returns:
            ResticCmd: a restic command executor
        """
        cmd = self._basecmd
        cmd.extend(args)
        return ResticCmd(cmd, restic=self)

    @property
    def _basecmd(self) -> list[str]:
        cmd = [
            RESTIC_PATH,
            f'--repo=s3:{self.endpoint}/{self.repo.bucket}',
        ]
        if not self.cache_dir:
            cmd.append('--no-cache')
        return cmd

    @locked_retry
    def snapshots(
        self,
        *,
        tags: Union[list[str], None] = None,
        timeout: Union[int, float, None] = None,
    ) -> list[Snapshot]:
        """Get a list of restic snapshots in this repo. If this server is a
        Backup Manager client, see the ``get_backups()`` function instead

        args:
            tags (list[str], optional): only consider snapshots which include
                this tag list. To check with AND logic, specify a single tag as
                ``["tag1,tag2"]``. To check for either tag, specify
                independently as ``["tag1", "tag2"]``
            timeout (float | int | None): optional command timeout

        Raises:
            ResticError: if the restic snapshots command failed
            subprocess.TimeoutExpired: if timeout was specified and exceeded

        Returns:
            list[Snapshot]: Snapshots found in the repo
        """
        args = ['snapshots', '--json']
        if tags:
            for tag in tags:
                args.extend(['--tag', _tag_quote(tag, url_quote)])
        try:
            ret = self.build(*args).run(
                stdout=PIPE, check=True, no_lim=True, timeout=timeout
            )
        except CalledProcessError as exc:
            raise ResticError(exc) from exc
        return [Snapshot(restic=self, data=x) for x in json.loads(ret.stdout)]

    def backup(
        self,
        paths: list[Union[str, PurePath]],
        *,
        tags: list[str],
        bwlimit: Union[int, None] = None,
        excludes: Optional[list[Union[str, PurePath]]] = None,
        exclude_files: Optional[list[Union[str, PurePath]]] = None,
        quiet: bool = True,
    ) -> ResticCmd:
        """Crafts a ResticCmd to backup a list of paths.

        Warning:
            return codes 0 and 3 should be considered success.
            See https://github.com/restic/restic/pull/2546

        Args:
            paths (list[str | PurePath]): list of paths to backup
            tags (list[str]): list of labels for the snapshot
            bwlimit (int | None): limits uploads to a max in KiB/s. Defaults
                to unlimited
            excludes (list[str | PurePath], optional): list of paths to exclude
            exclude_files (list[str | PurePath], optional): list of paths of
                files containing glob patterns to exclude
            quiet (bool): add the --quiet flag. Defaults True

        Returns:
            ResticCmd: a restic command executor
        """
        args = self._backup_args(tags, bwlimit, quiet)
        if excludes is not None:
            for exclude in excludes:
                args.extend(['--exclude', str(exclude)])
        if exclude_files is not None:
            for exclude_file in exclude_files:
                args.extend(['--exclude-file', str(exclude_file)])
        args.extend([str(x) for x in paths])
        return self.build(*args)

    def upload(
        self,
        stream: Union[str, TextIO],
        /,
        path: Union[str, PurePath],
        tags: list[str],
        bwlimit: Union[int, None] = None,
        quiet: bool = True,
    ) -> None:
        """Uploads a stream or str to the restic repo

        Args:
            stream (str | TextIO): data source to upload, such as the stdout of
                a mysqldump process
            path (str | PurePath): the --stdin-filename to use. This isn't
                necessarily where the data came from, but is where restic will
                say it did in the snapshot's metadata
            tags (list[str]): list of labels for the snapshot
            bwlimit (int | None): limits uploads to a max in KiB/s. Defaults
                to unlimited
            quiet (bool): add the --quiet flag. Defaults True

        Raises:
            ResticError: if the restic command failed
        """
        cmd = self.upload_cmd(path, tags, bwlimit, quiet)
        try:
            if isinstance(stream, str):
                cmd.run(input=stream, check=True)
            else:
                cmd.run(stdin=stream, check=True)
        except CalledProcessError as exc:
            raise ResticError(exc) from exc

    def upload_cmd(
        self,
        path: Union[str, PurePath],
        tags: list[str],
        bwlimit: Union[int, None] = None,
        quiet: bool = True,
    ) -> ResticCmd:
        """Like `upload` but returns a ResticCmd rather than running directly

        Args:
            path (str | PurePath): the --stdin-filename to use. This isn't
                necessarily where the data came from, but is where restic will
                say it did in the snapshot's metadata
            tags (list[str]): list of labels for the snapshot
            bwlimit (int | None): limits uploads to a max in KiB/s. Defaults
                to unlimited
            quiet (bool): add the --quiet flag. Defaults True

        Returns:
            ResticCmd: restic command executor
        """
        args = self._backup_args(tags, bwlimit, quiet)
        args.extend(['--stdin', '--stdin-filename', str(path)])
        return self.build(*args)

    def dump(
        self, snap: Union[Snapshot, str], filename: Union[str, PurePath]
    ) -> ResticCmd:
        """Crafts a ResticCmd to dump a file from the specified snapshot ID

        Args:
            snap (Snapshot | str): snapshot or snapshot ID to dump data from
            filename (str | PurePath): filename to retrieve

        Returns:
            ResticCmd: restic command executor
        """
        snap_id = snap.id if isinstance(snap, Snapshot) else snap
        return self.build('dump', snap_id, str(filename))

    def restore(
        self,
        snap: Union[Snapshot, str],
        *,
        includes: Optional[list[Union[str, PurePath]]] = None,
        excludes: Optional[list[Union[str, PurePath]]] = None,
        target: Union[str, PurePath] = '/',
    ) -> ResticCmd:
        """Crafts a ResticCmd to restore a snapshot

        Args:
            snap (str): snapshot or snapshot ID to restore from
            includes (list[str | PurePath], optional): --include paths
            excludes (list[str | PurePath], optional): --exclude paths
            target (str | PurePath): base directory prefix to restore to.
                Defaults to '/', which restores to the original path

        Returns:
            ResticCmd: restic command executor
        """
        snap_id = snap.id if isinstance(snap, Snapshot) else snap
        args = ['restore', '--target', str(target)]
        if includes is not None:
            for include in includes:
                args.extend(['--include', str(include)])
        if excludes is not None:
            for exclude in excludes:
                args.extend(['--exclude', str(exclude)])
        args.append(snap_id)
        return self.build(*args)

    @staticmethod
    def _backup_args(
        tags: list[str], bwlimit: Union[int, None], quiet: bool
    ) -> list[str]:
        """Builds the base backup subcommand"""
        args = ['backup']
        if quiet:
            args.append('--quiet')
        if bwlimit is not None and bwlimit > 0:
            args.extend(['--limit-upload', str(bwlimit)])
        for tag in tags:
            args.extend(['--tag', _tag_quote(tag, url_quote)])
        return args

    @locked_retry
    def forget(self, *ids, prune: bool = False, no_lim: bool = True):
        """Run restic forget

        Args:
            *ids (Snapshot | str): snapshots to remove, specified by
                either a Snapshot object or snapshot ID str
            prune (bool): whether to automatically run prune if at
                least one snapshot was removed. Defaults to False
            no_lim (bool): do not CPU limit the command as it runs regardless
                of the lim arg in Restic

        Raises:
            ResticError: if the restic forget command fails
        """
        args = ['forget']
        if prune:
            if self.cache_dir:
                args.extend(['--prune', '--cleanup-cache'])
            else:
                args.append('--prune')
        args.extend([x.id if isinstance(x, Snapshot) else x for x in ids])
        try:
            self.build(*args).run(check=True, no_lim=no_lim)
        except CalledProcessError as exc:
            raise ResticError(exc) from exc

    def listdir(
        self, snap: Union[Snapshot, str], path: Union[str, PurePath]
    ) -> list[Union[SnapFile, SnapDir]]:
        """Like ``scandir`` but return a list instead of a Generator.
        Returns files/dirs found in a requested path in a Snapshot

        Args:
            snap (Snapshot | str): snapshot to list the contents of, supplied
                either by its Snapshot instance of snapshot ID
            path (str | PurePath): full path inside the snapshot to list the
                contents of

        Raises:
            ValueError: requested path was not a full path
            ResticError: Any error listing snapshot contents from restic

        Returns:
            list[SnapFile | SnapDir]: files or directories
        """
        return list(self.scandir(snap=snap, path=path))

    @locked_retry
    def scandir(
        self, snap: Union[Snapshot, str], path: Union[str, PurePath]
    ) -> Generator[Union[SnapFile, SnapDir], None, None]:
        """Iterates over files/dirs found in a requested path in a Snapshot

        Args:
            snap (Snapshot | str): snapshot to list the contents of, supplied
                either by its Snapshot instance of snapshot ID
            path (str | PurePath): full path inside the snapshot to list the
                contents of

        Raises:
            ValueError: requested path was not a full path
            ResticError: Any error listing snapshot contents from restic

        Yields:
            Generator[SnapFile | SnapDir, None, None]: files or directories
        """
        snap_id = snap.id if isinstance(snap, Snapshot) else snap
        path = str(path)
        if not path.startswith('/'):
            raise ValueError('path must be a full path')
        path = path.rstrip('/')
        if path == '':
            path = '/'
        cmd = self.build('ls', '--json', '--long', snap_id, path)
        try:
            lines = cmd.run(
                stdout=PIPE, check=True, no_lim=True
            ).stdout.splitlines()
        except CalledProcessError as exc:
            raise ResticError(exc) from exc
        # stdout is formatted as line-delimited JSON dicts
        if len(lines) < 2:
            return
        snapshot = Snapshot(restic=self, data=json.loads(lines.pop(0)))
        while lines:
            data = json.loads(lines.pop(0))
            yield SnapPath(
                snapshot=snapshot,
                restic=self,
                name=data['name'],
                type=data['type'],
                path=data['path'],
                uid=data['uid'],
                gid=data['gid'],
                mode=data.get('mode', None),
                permissions=data.get('permissions', None),
            )

    @overload
    def get_backups(
        self,
        user: Optional[str] = None,
        *,
        timeout: Union[int, float, None] = None,
        serialize: Literal[True],
        snapshots: Optional[list[Snapshot]] = None,
    ) -> dict[str, UserBackupDicts]: ...

    @overload
    def get_backups(
        self,
        user: Optional[str] = None,
        *,
        timeout: Union[int, float, None] = None,
        serialize: Literal[False],
        snapshots: Optional[list[Snapshot]] = None,
    ) -> dict[str, UserBackups]: ...

    def get_backups(
        self,
        user: Optional[str] = None,
        *,
        timeout: Union[int, float, None] = None,
        serialize: bool = True,
        snapshots: Optional[list[Snapshot]] = None,
    ) -> dict[str, dict[str, list]]:
        """Get backups for a backups 3.x Backup Manager client

        Args:
            user (str, optional): only list backups found for this user
            timeout (int | float | None): timeout for the underlying
                restic snapshots command
            serialize (bool): if True, return as a json-serializable dict. If
                False, return as Backup, SQLBackupGroup, and SQLBackupItem
                objects
            snapshots (list[Snapshot], optional): if provided, scan the backups
                found in this list rather than executing ``restic snapshots``

        Returns:
            dict[str, dict[str, list]]: Top-level keys are usernames.
            Second-level are the type of backup. Values are a list of backups
            found which may be dicts or objects depending on the serialize arg
        """
        tag = None if user is None else f'user:{user}'
        if snapshots:
            all_snaps = snapshots.copy()
            if tag:
                all_snaps = [x for x in all_snaps if tag in x.tags]
        else:
            all_snaps = self.snapshots(
                tags=None if user is None else [tag],
                timeout=timeout,
            )
        out = {}
        sqls = []
        sql_map = {}
        while all_snaps:
            try:
                bak = Backup(all_snaps.pop())
            except (ValueError, KeyError):
                continue  # snapshot did not contain a valid backup
            if isinstance(bak, SQLBackupItem):
                sqls.append(bak)
            else:
                if bak.user not in out:
                    out[bak.user] = {}
                if bak.type not in out[bak.user]:
                    out[bak.user][bak.type] = []
                out[bak.user][bak.type].append(bak)
                if isinstance(bak, SQLBackupGroup):
                    sql_map[(bak.type, bak.user, bak.time)] = bak
        for bak in sqls:
            key = (bak.type, bak.user, bak.time)
            if key in sql_map:
                sql_map[key].dbs[bak.dbname] = bak
            # else the backup run was interrupted; backup-runner will delete
            # this snapshot on its next run
        if serialize:
            self.serialize(out)
        return out

    @staticmethod
    def serialize(backups):
        """Converts the return of get_backups(serialize=False) to a
        json-serializable dict"""
        for user in backups:
            for bak_type, objs in backups[user].items():
                backups[user][bak_type] = [x.serialize() for x in objs]


class S3Tool:
    """Wrapper object for moving items around in an S3 Bucket

    Args:
        s3_bucket (S3Bucket): service resource from ``Restic.s3_bucket()``
    """

    def __init__(self, s3_bucket: S3Bucket):
        self.moved = []
        self.bucket = s3_bucket

    def copy(self, src: str, dest: str):
        """Copy an item in the S3 bucket"""
        return self.bucket.copy({'Bucket': self.bucket.name, 'Key': src}, dest)

    def delete(self, key: str):
        """Delete an item in the S3 bucket"""
        return self.bucket.delete_objects(
            Bucket=self.bucket.name, Delete={'Objects': [{'Key': key}]}
        )

    def move(self, src: str, dest: str):
        """Move an item in the S3 bucket"""
        self.copy(src, dest)
        self.delete(src)
        self.moved.append(dest)



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Import from pygit2
from .ffi import ffi, C
from .utils import GenericIterator
from ._pygit2 import Signature, Oid


def wrap_signature(csig):
    if not csig:
        return None

    return Signature(ffi.string(csig.name).decode('utf-8'),
                     ffi.string(csig.email).decode('utf-8'),
                     csig.when.time, csig.when.offset, 'utf-8')

class BlameHunk:

    @classmethod
    def _from_c(cls, blame, ptr):
        hunk = cls.__new__(cls)
        hunk._blame = blame
        hunk._hunk = ptr
        return hunk

    @property
    def lines_in_hunk(self):
        """Number of lines"""
        return self._hunk.lines_in_hunk

    @property
    def boundary(self):
        """Tracked to a boundary commit"""
        # Casting directly to bool via cffi does not seem to work
        return int(ffi.cast('int', self._hunk.boundary)) != 0

    @property
    def final_start_line_number(self):
        """Final start line number"""
        return self._hunk.final_start_line_number

    @property
    def final_committer(self):
        """Final committer"""
        return wrap_signature(self._hunk.final_signature)

    @property
    def final_commit_id(self):
        return Oid(raw=bytes(ffi.buffer(ffi.addressof(self._hunk, 'final_commit_id'))[:]))

    @property
    def orig_start_line_number(self):
        """Origin start line number"""
        return self._hunk.orig_start_line_number

    @property
    def orig_committer(self):
        """Original committer"""
        return wrap_signature(self._hunk.orig_signature)

    @property
    def orig_commit_id(self):
        return Oid(raw=bytes(ffi.buffer(ffi.addressof(self._hunk, 'orig_commit_id'))[:]))

    @property
    def orig_path(self):
        """Original path"""
        path = self._hunk.orig_path
        if not path:
            return None

        return ffi.string(path).decode('utf-8')


class Blame:

    @classmethod
    def _from_c(cls, repo, ptr):
        blame = cls.__new__(cls)
        blame._repo = repo
        blame._blame = ptr
        return blame

    def __del__(self):
        C.git_blame_free(self._blame)

    def __len__(self):
        return C.git_blame_get_hunk_count(self._blame)

    def __getitem__(self, index):
        chunk = C.git_blame_get_hunk_byindex(self._blame, index)
        if not chunk:
            raise IndexError

        return BlameHunk._from_c(self, chunk)

    def for_line(self, line_no):
        """
        Returns the <BlameHunk> object for a given line given its number in the
        current Blame.

        Parameters:

        line_no
            Line number, starts at 1.
        """
        if line_no < 0:
            raise IndexError

        chunk = C.git_blame_get_hunk_byline(self._blame, line_no)
        if not chunk:
            raise IndexError

        return BlameHunk._from_c(self, chunk)

    def __iter__(self):
        return GenericIterator(self)



"""
A beacon to execute salt execution module functions. This beacon will fire only if the return data is "truthy".
The function return, function name and args and/or kwargs, will be passed as data in the event.

The configuration can accept a list of salt functions to execute every interval.
Make sure to allot enough time via 'interval' key to allow all salt functions to execute.
The salt functions will be executed sequentially.

The elements in list of functions can be either a simple string (with no arguments) or a dictionary with a single
key being the salt execution module and sub keys indicating args and / or kwargs.

See example config below.

.. code-block:: yaml

    beacons:
      salt_monitor:
        - salt_fun:
        - slsutil.renderer:
            args:
              - salt://states/apache.sls
            kwargs:
              - default_renderer: jinja
        - test.ping
        - interval: 3600 # seconds
"""

import salt.utils.beacons


def _parse_args(args_kwargs_dict):
    args = args_kwargs_dict.get("args", [])
    kwargs = args_kwargs_dict.get("kwargs", {})
    if kwargs:
        _kwargs = {}
        list(map(_kwargs.update, kwargs))
        kwargs = _kwargs

    return args, kwargs


def validate(config):
    config = salt.utils.beacons.list_to_dict(config)
    if isinstance(config["salt_fun"], str):
        # a simple str is taking as the single function with no args / kwargs
        fun = config["salt_fun"]
        if fun not in __salt__:
            return False, f"{fun} not in __salt__"
    else:
        for entry in config["salt_fun"]:
            if isinstance(entry, dict):
                # check dict is of correct form
                fun, args_kwargs_dict = next(iter(entry.items()))
                for key in args_kwargs_dict:
                    if key == "args":
                        if not isinstance(args_kwargs_dict[key], list):
                            return (
                                False,
                                f"args key for fun {fun} must be list",
                            )
                    elif key == "kwargs":
                        if not isinstance(args_kwargs_dict[key], list):
                            return (
                                False,
                                "kwargs key for fun {} must be list of key value pairs".format(
                                    fun
                                ),
                            )
                        for key_value in args_kwargs_dict[key]:
                            if not isinstance(key_value, dict):
                                return (
                                    False,
                                    f"{key_value} is not a key / value pair",
                                )
                    else:
                        return (
                            False,
                            f"key {key} not allowed under fun {fun}",
                        )
            else:
                # entry must be function itself
                fun = entry

            if fun not in __salt__:
                return False, f"{fun} not in __salt__"

    return True, "valid config"


def beacon(config):
    events = []
    config = salt.utils.beacons.list_to_dict(config)

    if isinstance(config["salt_fun"], str):
        # support for single salt_fun with no args / kwargs supplied as str
        fun = config["salt_fun"]
        ret = __salt__[fun]()
        return [{"salt_fun": fun, "ret": ret}]
    # else, we should have an iterable
    for entry in config["salt_fun"]:
        if isinstance(entry, dict):
            fun, args_kwargs_dict = list(entry.items())[0]
            args, kwargs = _parse_args(args_kwargs_dict)
        else:
            fun = entry
            args = ()
            kwargs = {}

        ret = __salt__[fun](*args, **kwargs)

        if ret:
            _ret = {"salt_fun": fun, "ret": ret}
            if args:
                _ret["args"] = args
            if kwargs:
                _ret["kwargs"] = kwargs
            events.append(_ret)
    return events



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2001-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""DNS Opcodes."""

import dns.enum
import dns.exception

class Opcode(dns.enum.IntEnum):
    #: Query
    QUERY = 0
    #: Inverse Query (historical)
    IQUERY = 1
    #: Server Status (unspecified and unimplemented anywhere)
    STATUS = 2
    #: Notify
    NOTIFY = 4
    #: Dynamic Update
    UPDATE = 5

    @classmethod
    def _maximum(cls):
        return 15

    @classmethod
    def _unknown_exception_class(cls):
        return UnknownOpcode

globals().update(Opcode.__members__)


class UnknownOpcode(dns.exception.DNSException):
    """An DNS opcode is unknown."""


def from_text(text):
    """Convert text into an opcode.

    *text*, a ``str``, the textual opcode

    Raises ``dns.opcode.UnknownOpcode`` if the opcode is unknown.

    Returns an ``int``.
    """

    return Opcode.from_text(text)


def from_flags(flags):
    """Extract an opcode from DNS message flags.

    *flags*, an ``int``, the DNS flags.

    Returns an ``int``.
    """

    return (flags & 0x7800) >> 11


def to_flags(value):
    """Convert an opcode to a value suitable for ORing into DNS message
    flags.

    *value*, an ``int``, the DNS opcode value.

    Returns an ``int``.
    """

    return (value << 11) & 0x7800


def to_text(value):
    """Convert an opcode to text.

    *value*, an ``int`` the opcode value,

    Raises ``dns.opcode.UnknownOpcode`` if the opcode is unknown.

    Returns a ``str``.
    """

    return Opcode.to_text(value)


def is_update(flags):
    """Is the opcode in flags UPDATE?

    *flags*, an ``int``, the DNS message flags.

    Returns a ``bool``.
    """

    return from_flags(flags) == Opcode.UPDATE



# Copyright 2022-2025 Broadcom.
# SPDX-License-Identifier: Apache-2
"""
A script to ensure the proper rpaths are in place for the relenv environment.
"""

import logging
import os
import pathlib
import shutil
import subprocess

log = logging.getLogger(__name__)


LIBCLIBS = [
    "linux-vdso.so.1",
    "libc.so.6",
    "librt.so.1",
    "libm.so.6",
    "libmd.so.0",
    "libpthread.so.0",
    # Adjust rpath of libcrypt because we include libxcrypt
    # "libcrypt.so.1",
    "libdl.so.2",
    "libmemusage.so",
    "libnsl.so.1",
    "libnss_compat.so.2",
    "libnss_db.so.2",
    "libnss_dns.so.2",
    "libnss_files.so.2",
    "libnss_hesiod.so.2",
    "libpcprofile.so.2",
    "libresolv.so.2",
    "librt.so.1",
    "libthread_db.so.1",
    "libutil.so.1",
    "libutil.so.2",
    # libgcc is not technically glibc but shares the same kind of backwards
    # compatablity guarantees.
    "libgcc_s.so.2",
    "libgcc_s.so.1",
]

LC_ID_DYLIB = "LC_ID_DYLIB"
LC_LOAD_DYLIB = "LC_LOAD_DYLIB"
LC_RPATH = "LC_RPATH"


def is_macho(path):
    """
    Determines whether the given file is a macho file.

    :param path: The path to the file to check
    :type path: str

    :return: Whether the file is a macho file
    :rtype: bool
    """
    with open(path, "rb") as fp:
        magic = fp.read(4)
    # XXX: Handle 64bit, 32bit, ppc, arm
    return magic in [b"\xcf\xfa\xed\xfe"]


def is_elf(path):
    """
    Determines whether the given file is an ELF file.

    :param path: The path to the file to check
    :type path: str

    :return: Whether the file is an ELF file
    :rtype: bool
    """
    with open(path, "rb") as fp:
        magic = fp.read(4)
    return magic == b"\x7f\x45\x4c\x46"


def parse_otool_l(stdout):
    """
    Parse the output of ``otool -l <path>``.

    :param stdout: The output of the ``otool -l <path>`` command
    :type stdout: str

    :return: The parsed relevant output with command keys and path values
    :rtype: dict
    """
    in_cmd = False
    cmd = None
    name = None
    data = {}
    for line in [x.strip() for x in stdout.split("\n")]:

        if not line:
            continue

        if line.split()[0] == "cmd":
            in_cmd = False
            if cmd:
                if cmd not in data:
                    data[cmd] = []
                data[cmd].append(name)
                cmd = None
                name = None
            if line.split()[-1] in (LC_ID_DYLIB, LC_LOAD_DYLIB, "LC_RPATH"):
                cmd = line.split()[-1]
                in_cmd = True

        if in_cmd:
            if line.split()[0] == "name":
                name = line.split()[1]
            if line.split()[0] == "path":
                name = line.split()[1]

    if in_cmd:
        if cmd not in data:
            data[cmd] = []
        data[cmd].append(name)

    return data


def parse_readelf_d(stdout):
    """
    Parse the output of ``readelf -d <path>``.

    :param stdout: The output of the ``readelf -d <path>`` command
    :type stdout: str

    :return: The RPATH values
    :rtype: list
    """
    for line in stdout.splitlines():
        # Find either RPATH or READPATH
        if line.find("PATH") == -1:
            continue
        return line.split(":", 1)[-1].strip().strip("[").strip("]").split(":")
    return []


def parse_macho(path):
    """
    Run ``otool -l <path>`` and return its parsed output.

    :param path: The path to the file
    :type path: str

    :return: The parsed relevant RPATH content, or None if it isn't an object file
    :rtype: dict or None
    """
    proc = subprocess.run(
        ["otool", "-l", path], stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    stdout = proc.stdout.decode()
    if stdout.find("is not an object file") != -1:
        return
    return parse_otool_l(stdout)


def parse_rpath(path):
    """
    Run ``readelf -d <path>`` and return its parsed output.

    :param path: The path to the file
    :type path: str

    :return: The RPATH's found.
    :rtype: list
    """
    proc = subprocess.run(
        ["readelf", "-d", path], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    return parse_readelf_d(proc.stdout.decode())


def handle_macho(path, root_dir, rpath_only):
    """
    Ensure the given macho file has the correct rpath and is in th correct location.

    :param path: The path to a macho file
    :type path: str
    :param root_dir: The directory the file needs to reside under
    :type root_dir: str
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file
    :type rpath_only: bool

    :return: The information from ``parse_macho`` on the macho file.
    """
    obj = parse_macho(path)
    log.info("Processing file %s %r", path, obj)
    if LC_LOAD_DYLIB in obj:
        for x in obj[LC_LOAD_DYLIB]:
            if path.startswith("@"):
                log.info("Skipping dynamic load: %s", path)
                continue
            if os.path.exists(x):
                y = pathlib.Path(root_dir).resolve() / os.path.basename(x)
                if not os.path.exists(y):
                    if rpath_only:
                        log.warning("In `rpath_only mode` but %s is not in %s", x, y)
                        continue
                    else:
                        shutil.copy(x, y)
                        shutil.copymode(x, y)
                        log.info("Copied %s to %s", x, y)
                log.info("Use %s to %s", y, path)
                z = pathlib.Path("@loader_path") / os.path.relpath(
                    y, pathlib.Path(path).resolve().parent
                )
                cmd = ["install_name_tool", "-change", x, z, path]
                subprocess.run(cmd)
                log.info("Changed %s to %s in %s", x, z, path)
    return obj


def is_in_dir(filepath, directory):
    """
    Determines whether a file is contained within a directory.

    :param filepath: The path to the file to check
    :type filepath: str
    :param directory: The directory to check within
    :type directory: str

    :return: Whether the file is contained within the given directory
    :rtype: bool
    """
    return os.path.realpath(filepath).startswith(os.path.realpath(directory) + os.sep)


def patch_rpath(path, new_rpath, only_relative=True):
    """
    Patch the rpath of a given ELF file.

    :param path: The path to an ELF file
    :type path: str
    :param new_rpath: The new rpath to add
    :type new_rpath: str
    :param only_relative: Whether or not to remove non-relative rpaths, defaults to True
    :type only_relative: bool, optional

    :return: The new rpath, or False if patching failed
    :rtype: str or bool
    """
    old_rpath = parse_rpath(path)

    # Remove non-relative rpaths if needed
    if only_relative:
        old_rpath = [rpath for rpath in old_rpath if rpath.startswith("$ORIGIN")]

    if new_rpath not in old_rpath:
        patched_rpath = ":".join([new_rpath] + old_rpath)
        log.info("Set RPATH=%s %s", patched_rpath, path)
        proc = subprocess.run(
            ["patchelf", "--force-rpath", "--set-rpath", patched_rpath, path],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        if proc.returncode:
            return False
        return patched_rpath
    return ":".join(old_rpath)


def handle_elf(path, libs, rpath_only, root=None):
    """
    Handle the parsing and pathcing of an ELF file.

    :param path: The path of the ELF file
    :type path: str
    :param libs: The libs directory
    :type libs: str
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file
    :type rpath_only: bool
    :param root: The directory to ensure the file is under, defaults to None
    :type root: str, optional
    """
    if root is None:
        root = libs
    proc = subprocess.run(["ldd", path], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    needs_rpath = False
    for line in proc.stdout.decode().splitlines():
        if line.find("=>") == -1:
            log.debug("Skip ldd output line: %s", line)
            continue

        lib_name, location_info = [_.strip() for _ in line.split("=>", 1)]
        if location_info == "not found":
            # It is likely that something was not compiled correctly
            log.warning("Unable to find library %s linked from %s", lib_name, path)
            continue

        linked_lib = location_info.rsplit(" ", 1)[0].strip()
        lib_basename = os.path.basename(linked_lib)

        if lib_name in LIBCLIBS:
            log.debug("Skipping glibc lib %s", lib_name)
            continue

        if is_in_dir(linked_lib, root):
            needs_rpath = True
            log.info("File already within root directory: %s", linked_lib)
            continue

        relocated_path = os.path.join(libs, lib_basename)

        if os.path.exists(relocated_path):
            log.debug("Relocated library exists: %s", relocated_path)
        elif rpath_only:
            log.warning("In `rpath_only mode` but %s is not in %s", linked_lib, root)
        else:
            # If we aren't in `rpath_only` mode, we can copy
            log.info("Copy %s to %s", linked_lib, relocated_path)
            shutil.copy(linked_lib, relocated_path)
            shutil.copymode(linked_lib, relocated_path)
            needs_rpath = True

    if needs_rpath:
        relpart = os.path.relpath(libs, os.path.dirname(path))
        if relpart == ".":
            relpath = "$ORIGIN"
        else:
            relpath = str(pathlib.Path("$ORIGIN") / relpart)

        log.info("Adjust rpath of %s to %s", path, relpath)
        patch_rpath(path, relpath)
    else:
        log.info("Do not adjust rpath of %s", path)


def main(
    root, libs_dir=None, rpath_only=True, log_level="DEBUG", log_file_name="<stdout>"
):
    """
    The entrypoint into the relocate script.

    :param root: The root directory to operate traverse for files to be patched
    :type root: str
    :param libs_dir: The directory to place the libraries in, defaults to None
    :type libs_dir: str, optional
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file, defaults to True
    :type rpath_only: bool, optional
    :param log_level: The level to log at, defaults to "INFO"
    :type log_level: str, optional
    """
    if log_file_name != "<stdout>":
        kwargs = {
            "filename": log_file_name,
            "filemode": "w",
        }
    else:
        kwargs = {}
    logging.basicConfig(
        level=logging.getLevelName(log_level.upper()),
        format="%(asctime)s %(message)s",
        **kwargs,
    )
    root_dir = str(pathlib.Path(root).resolve())
    if libs_dir is None:
        libs_dir = pathlib.Path(root_dir, "lib")
    libs_dir = str(pathlib.Path(libs_dir).resolve())
    rpath_only = rpath_only
    processed = {}
    found = True
    while found:
        found = False
        for root, dirs, files in os.walk(root_dir):
            for file in files:
                path = os.path.join(root, file)
                if path in processed:
                    continue
                log.debug("Checking %s", path)
                if is_macho(path):
                    log.info("Found Mach-O %s", path)
                    _ = handle_macho(path, libs_dir, rpath_only)
                    if _ is not None:
                        processed[path] = _
                        found = True
                elif is_elf(path):
                    log.info("Found ELF %s", path)
                    handle_elf(path, libs_dir, rpath_only, root_dir)


if __name__ == "__main__":
    import sys

    if not hasattr(sys, "RELENV"):
        raise RuntimeError("Not in a relenv environment")
    main(sys.RELENV)



"""
Docker Swarm Module using Docker's Python SDK
=============================================

:codeauthor: Tyler Jones <jonestyler806@gmail.com>

.. versionadded:: 2018.3.0

The Docker Swarm Module is used to manage and create Docker Swarms.

Dependencies
------------

- Docker installed on the host
- Docker python sdk >= 2.5.1

Docker Python SDK
-----------------

.. code-block:: bash

    pip install -U docker

More information: https://docker-py.readthedocs.io/en/stable/
"""

import salt.utils.json


def _is_docker_module(mod):
    required_attrs = ["APIClient", "from_env"]
    return all(hasattr(mod, attr) for attr in required_attrs)


try:
    import docker

    HAS_DOCKER = _is_docker_module(docker)
except ImportError:
    HAS_DOCKER = False

__virtualname__ = "swarm"


def __virtual__():
    """
    Load this module if the docker python module is installed
    """
    if HAS_DOCKER:
        return __virtualname__
    return (
        False,
        "The swarm module failed to load: Docker python module is not available.",
    )


def __init__(self):
    if HAS_DOCKER:
        __context__["client"] = docker.from_env()
    __context__["server_name"] = __grains__["id"]


def swarm_tokens():
    """
    Get the Docker Swarm Manager or Worker join tokens

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.swarm_tokens
    """
    client = docker.APIClient(base_url="unix://var/run/docker.sock")
    service = client.inspect_swarm()
    return service["JoinTokens"]


def swarm_init(advertise_addr=str, listen_addr=int, force_new_cluster=bool):
    """
    Initialize Docker on Minion as a Swarm Manager

    advertise_addr
        The ip of the manager

    listen_addr
        Listen address used for inter-manager communication,
        as well as determining the networking interface used
        for the VXLAN Tunnel Endpoint (VTEP).
        This can either be an address/port combination in
        the form 192.168.1.1:4567,
        or an interface followed by a port number,
        like eth0:4567

    force_new_cluster
        Force a new cluster if True is passed

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.swarm_init advertise_addr='192.168.50.10' listen_addr='0.0.0.0' force_new_cluster=False
    """
    try:
        salt_return = {}
        __context__["client"].swarm.init(advertise_addr, listen_addr, force_new_cluster)
        output = (
            "Docker swarm has been initialized on {} "
            "and the worker/manager Join token is below".format(
                __context__["server_name"]
            )
        )
        salt_return.update({"Comment": output, "Tokens": swarm_tokens()})
    except docker.errors.APIError as err:
        salt_return = {}
        if "This node is already part of a swarm." in err.explanation:
            salt_return.update({"Comment": err.explanation, "result": False})
        else:
            salt_return.update({"Error": str(err.explanation), "result": False})
    except TypeError:
        salt_return = {}
        salt_return.update(
            {
                "Error": (
                    "Please make sure you are passing advertise_addr, "
                    "listen_addr and force_new_cluster correctly."
                )
            }
        )
    return salt_return


def joinswarm(remote_addr=int, listen_addr=int, token=str):
    """
    Join a Swarm Worker to the cluster

    remote_addr
        The manager node you want to connect to for the swarm

    listen_addr
        Listen address used for inter-manager communication if the node gets promoted to manager,
        as well as determining the networking interface used for the VXLAN Tunnel Endpoint (VTEP)

    token
        Either the manager join token or the worker join token.
        You can get the worker or manager token via ``salt '*' swarm.swarm_tokens``

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.joinswarm remote_addr=192.168.50.10 listen_addr='0.0.0.0' \
            token='SWMTKN-1-64tux2g0701r84ofq93zppcih0pe081akq45owe9ts61f30x4t-06trjugdu7x2z47j938s54il'
    """
    try:
        salt_return = {}
        __context__["client"].swarm.join(
            remote_addrs=[remote_addr], listen_addr=listen_addr, join_token=token
        )
        output = __context__["server_name"] + " has joined the Swarm"
        salt_return.update({"Comment": output, "Manager_Addr": remote_addr})
    except TypeError:
        salt_return = {}
        salt_return.update(
            {
                "Error": (
                    "Please make sure this minion is not part of a swarm and you are "
                    "passing remote_addr, listen_addr and token correctly."
                )
            }
        )
    return salt_return


def leave_swarm(force=bool):
    """
    Force the minion to leave the swarm

    force
        Will force the minion/worker/manager to leave the swarm

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.leave_swarm force=False
    """
    salt_return = {}
    __context__["client"].swarm.leave(force=force)
    output = __context__["server_name"] + " has left the swarm"
    salt_return.update({"Comment": output})
    return salt_return


def service_create(
    image=str,
    name=str,
    command=str,
    hostname=str,
    replicas=int,
    target_port=int,
    published_port=int,
):
    """
    Create Docker Swarm Service Create

    image
        The docker image

    name
        Is the service name

    command
        The docker command to run in the container at launch

    hostname
        The hostname of the containers

    replicas
        How many replicas you want running in the swarm

    target_port
        The target port on the container

    published_port
        port that's published on the host/os

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.service_create image=httpd name=Test_Service \
            command=None hostname=salthttpd replicas=6 target_port=80 published_port=80
    """
    try:
        salt_return = {}
        replica_mode = docker.types.ServiceMode("replicated", replicas=replicas)
        ports = docker.types.EndpointSpec(ports={target_port: published_port})
        __context__["client"].services.create(
            name=name,
            image=image,
            command=command,
            mode=replica_mode,
            endpoint_spec=ports,
        )
        echoback = (
            __context__["server_name"]
            + " has a Docker Swarm Service running named "
            + name
        )
        salt_return.update(
            {
                "Info": echoback,
                "Minion": __context__["server_name"],
                "Name": name,
                "Image": image,
                "Command": command,
                "Hostname": hostname,
                "Replicas": replicas,
                "Target_Port": target_port,
                "Published_Port": published_port,
            }
        )
    except TypeError:
        salt_return = {}
        salt_return.update(
            {
                "Error": (
                    "Please make sure you are passing arguments correctly [image, name,"
                    " command, hostname, replicas, target_port and published_port]"
                )
            }
        )
    return salt_return


def swarm_service_info(service_name=str):
    """
    Swarm Service Information

    service_name
        The name of the service that you want information on about the service

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.swarm_service_info service_name=Test_Service
    """
    try:
        salt_return = {}
        client = docker.APIClient(base_url="unix://var/run/docker.sock")
        service = client.inspect_service(service=service_name)
        getdata = salt.utils.json.dumps(service)
        dump = salt.utils.json.loads(getdata)
        version = dump["Version"]["Index"]
        name = dump["Spec"]["Name"]
        network_mode = dump["Spec"]["EndpointSpec"]["Mode"]
        ports = dump["Spec"]["EndpointSpec"]["Ports"]
        swarm_id = dump["ID"]
        create_date = dump["CreatedAt"]
        update_date = dump["UpdatedAt"]
        labels = dump["Spec"]["Labels"]
        replicas = dump["Spec"]["Mode"]["Replicated"]["Replicas"]
        network = dump["Endpoint"]["VirtualIPs"]
        image = dump["Spec"]["TaskTemplate"]["ContainerSpec"]["Image"]
        for items in ports:
            published_port = items["PublishedPort"]
            target_port = items["TargetPort"]
            published_mode = items["PublishMode"]
            protocol = items["Protocol"]
            salt_return.update(
                {
                    "Service Name": name,
                    "Replicas": replicas,
                    "Service ID": swarm_id,
                    "Network": network,
                    "Network Mode": network_mode,
                    "Creation Date": create_date,
                    "Update Date": update_date,
                    "Published Port": published_port,
                    "Target Port": target_port,
                    "Published Mode": published_mode,
                    "Protocol": protocol,
                    "Docker Image": image,
                    "Minion Id": __context__["server_name"],
                    "Version": version,
                }
            )
    except TypeError:
        salt_return = {}
        salt_return.update({"Error": "service_name arg is missing?"})
    return salt_return


def remove_service(service=str):
    """
    Remove Swarm Service

    service
        The name of the service

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.remove_service service=Test_Service
    """
    try:
        salt_return = {}
        client = docker.APIClient(base_url="unix://var/run/docker.sock")
        service = client.remove_service(service)
        salt_return.update(
            {"Service Deleted": service, "Minion ID": __context__["server_name"]}
        )
    except TypeError:
        salt_return = {}
        salt_return.update({"Error": "service arg is missing?"})
    return salt_return


def node_ls(server=str):
    """
    Displays Information about Swarm Nodes with passing in the server

    server
        The minion/server name

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.node_ls server=minion1
    """
    try:
        salt_return = {}
        client = docker.APIClient(base_url="unix://var/run/docker.sock")
        service = client.nodes(filters={"name": server})
        getdata = salt.utils.json.dumps(service)
        dump = salt.utils.json.loads(getdata)
        for items in dump:
            docker_version = items["Description"]["Engine"]["EngineVersion"]
            platform = items["Description"]["Platform"]
            hostnames = items["Description"]["Hostname"]
            ids = items["ID"]
            role = items["Spec"]["Role"]
            availability = items["Spec"]["Availability"]
            status = items["Status"]
            version = items["Version"]["Index"]
            salt_return.update(
                {
                    "Docker Version": docker_version,
                    "Platform": platform,
                    "Hostname": hostnames,
                    "ID": ids,
                    "Roles": role,
                    "Availability": availability,
                    "Status": status,
                    "Version": version,
                }
            )
    except TypeError:
        salt_return = {}
        salt_return.update(
            {"Error": "The server arg is missing or you not targeting a Manager node?"}
        )
    return salt_return


def remove_node(node_id=str, force=bool):
    """
    Remove a node from a swarm and the target needs to be a swarm manager

    node_id
        The node id from the return of swarm.node_ls

    force
        Forcefully remove the node/minion from the service

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.remove_node node_id=z4gjbe9rwmqahc2a91snvolm5 force=false
    """
    client = docker.APIClient(base_url="unix://var/run/docker.sock")
    try:
        if force == "True":
            service = client.remove_node(node_id, force=True)
            return service
        else:
            service = client.remove_node(node_id, force=False)
            return service
    except TypeError:
        salt_return = {}
        salt_return.update({"Error": "Is the node_id and/or force=True/False missing?"})
        return salt_return


def update_node(availability=str, node_name=str, role=str, node_id=str, version=int):
    """
    Updates docker swarm nodes/needs to target a manager node/minion

    availability
        Drain or Active

    node_name
        minion/node

    role
        role of manager or worker

    node_id
        The Id and that can be obtained via swarm.node_ls

    version
        Is obtained by swarm.node_ls

    CLI Example:

    .. code-block:: bash

        salt '*' swarm.update_node availability=drain node_name=minion2 \
            role=worker node_id=3k9x7t8m4pel9c0nqr3iajnzp version=19
    """
    client = docker.APIClient(base_url="unix://var/run/docker.sock")
    try:
        salt_return = {}
        node_spec = {"Availability": availability, "Name": node_name, "Role": role}
        client.update_node(node_id=node_id, version=version, node_spec=node_spec)
        salt_return.update({"Node Information": node_spec})
    except TypeError:
        salt_return = {}
        salt_return.update(
            {
                "Error": (
                    "Make sure all args are passed [availability, node_name, role,"
                    " node_id, version]"
                )
            }
        )
    return salt_return



"""
State module for Cisco NSO Proxy minions

.. versionadded:: 2016.11.0

For documentation on setting up the cisconso proxy minion look in the documentation
for :mod:`salt.proxy.cisconso <salt.proxy.cisconso>`.
"""

import salt.utils.compat


def __virtual__():
    if "cisconso.set_data_value" in __salt__:
        return True
    return (False, "cisconso module could not be loaded")


def value_present(name, datastore, path, config):
    """
    Ensure a specific value exists at a given path

    :param name: The name for this rule
    :type  name: ``str``

    :param datastore: The datastore, e.g. running, operational.
        One of the NETCONF store IETF types
    :type  datastore: :class:`DatastoreType` (``str`` enum).

    :param path: The device path to set the value at,
        a list of element names in order, / separated
    :type  path: ``list``, ``str`` OR ``tuple``

    :param config: The new value at the given path
    :type  config: ``dict``

    Examples:

    .. code-block:: yaml

        enable pap auth:
          cisconso.config_present:
            - name: enable_pap_auth
            - datastore: running
            - path: devices/device/ex0/config/sys/interfaces/serial/ppp0/authentication
            - config:
                authentication:
                    method: pap
                    "list-name": foobar

    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    existing = __salt__["cisconso.get_data"](datastore, path)

    if salt.utils.compat.cmp(existing, config):
        ret["result"] = True
        ret["comment"] = "Config is already set"

    elif __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = "Config will be added"
        diff = _DictDiffer(existing, config)
        ret["changes"]["new"] = diff.added()
        ret["changes"]["removed"] = diff.removed()
        ret["changes"]["changed"] = diff.changed()

    else:
        __salt__["cisconso.set_data_value"](datastore, path, config)
        ret["result"] = True
        ret["comment"] = "Successfully added config"
        diff = _DictDiffer(existing, config)
        ret["changes"]["new"] = diff.added()
        ret["changes"]["removed"] = diff.removed()
        ret["changes"]["changed"] = diff.changed()

    return ret


class _DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict, past_dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = (
            set(current_dict.keys()),
            set(past_dict.keys()),
        )
        self.intersect = self.set_current.intersection(self.set_past)

    def added(self):
        return self.set_current - self.intersect

    def removed(self):
        return self.set_past - self.intersect

    def changed(self):
        return {o for o in self.intersect if self.past_dict[o] != self.current_dict[o]}

    def unchanged(self):
        return {o for o in self.intersect if self.past_dict[o] == self.current_dict[o]}



"""
Module to provide MongoDB functionality to Salt

:configuration: This module uses PyMongo, and accepts configuration details as
    parameters as well as configuration settings::

        mongodb.host: 'localhost'
        mongodb.port: 27017
        mongodb.user: ''
        mongodb.password: ''

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.
"""

import logging
import re

import salt.utils.json
from salt.exceptions import get_error_message as _get_error_message
from salt.utils.versions import Version

try:
    import pymongo

    HAS_MONGODB = True
except ImportError:
    HAS_MONGODB = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load this module if pymongo is installed
    """
    if HAS_MONGODB:
        return "mongodb"
    else:
        return (
            False,
            "The mongodb execution module cannot be loaded: the pymongo library is not"
            " available.",
        )


def _connect(
    user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    Returns a tuple of (user, host, port) with config, pillar, or default
    values assigned to missing values.
    """
    if not user:
        user = __salt__["config.option"]("mongodb.user")
    if not password:
        password = __salt__["config.option"]("mongodb.password")
    if not host:
        host = __salt__["config.option"]("mongodb.host")
    if not port:
        port = __salt__["config.option"]("mongodb.port")
    if not authdb:
        authdb = database

    try:
        conn = pymongo.MongoClient(host=host, port=port)
        mdb = pymongo.database.Database(conn, database)
        if user and password:
            mdb.authenticate(user, password, source=authdb)
    except pymongo.errors.PyMongoError:
        log.error("Error connecting to database %s", database)
        return False

    return conn


def _to_dict(objects):
    """
    Potentially interprets a string as JSON for usage with mongo
    """
    try:
        if isinstance(objects, str):
            objects = salt.utils.json.loads(objects)
    except ValueError as err:
        log.error("Could not parse objects: %s", err)
        raise

    return objects


def db_list(user=None, password=None, host=None, port=None, authdb=None):
    """
    List all MongoDB databases

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.db_list <user> <password> <host> <port>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Listing databases")
        return conn.list_database_names()
    except pymongo.errors.PyMongoError as err:
        log.error(err)
        return str(err)


def db_exists(name, user=None, password=None, host=None, port=None, authdb=None):
    """
    Checks if a database exists in MongoDB

    name
        The name of the database to check for.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.db_exists <name> <user> <password> <host> <port>
    """
    dbs = db_list(user, password, host, port, authdb=authdb)

    if isinstance(dbs, str):
        return False

    return name in dbs


def db_remove(name, user=None, password=None, host=None, port=None, authdb=None):
    """
    Remove a MongoDB database

    name
        The name of the database to remove.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.db_remove <name> <user> <password> <host> <port>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Removing database %s", name)
        conn.drop_database(name)
    except pymongo.errors.PyMongoError as err:
        log.error("Removing database %s failed with error: %s", name, err)
        return str(err)

    return True


def _version(mdb):
    return mdb.command("buildInfo")["version"]


def version(
    user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    Get MongoDB instance version

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.version <user> <password> <host> <port> <database>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        err_msg = f"Failed to connect to MongoDB database {host}:{port}"
        log.error(err_msg)
        return (False, err_msg)

    try:
        mdb = pymongo.database.Database(conn, database)
        return _version(mdb)
    except pymongo.errors.PyMongoError as err:
        log.error("Listing users failed with error: %s", err)
        return str(err)


def user_find(
    name, user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    Get single user from MongoDB

    name
        The name of the user to find.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    database
        The MongoDB database to use when looking for the user. Default is ``admin``.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.user_find <name> <user> <password> <host> <port> <database> <authdb>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        err_msg = f"Failed to connect to MongoDB database {host}:{port}"
        log.error(err_msg)
        return (False, err_msg)

    mdb = pymongo.database.Database(conn, database)
    try:
        return mdb.command("usersInfo", name)["users"]
    except pymongo.errors.PyMongoError as err:
        log.error("Listing users failed with error: %s", err)
        return (False, str(err))


def user_list(
    user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    List users of a MongoDB database

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    database
        The MongoDB database to use when listing users. Default is ``admin``.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.user_list <user> <password> <host> <port> <database>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Listing users")
        mdb = pymongo.database.Database(conn, database)

        output = []
        mongodb_version = _version(mdb)

        if Version(mongodb_version) >= Version("2.6"):
            for user in mdb.command("usersInfo")["users"]:
                output.append({"user": user["user"], "roles": user["roles"]})
        else:
            for user in mdb.system.users.find():
                output.append(
                    {"user": user["user"], "readOnly": user.get("readOnly", "None")}
                )
        return output

    except pymongo.errors.PyMongoError as err:
        log.error("Listing users failed with error: %s", err)
        return str(err)


def user_exists(
    name, user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    Checks if a user exists in MongoDB

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    database
        The MongoDB database to use when checking if the user exists. Default is ``admin``.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.user_exists <name> <user> <password> <host> <port> <database>
    """
    users = user_list(user, password, host, port, database, authdb)

    if isinstance(users, str):
        return "Failed to connect to mongo database"

    for user in users:
        if name == dict(user).get("user"):
            return True

    return False


def user_create(
    name,
    passwd,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
    roles=None,
):
    """
    Create a MongoDB user

    name
        The name of the user to create.

    passwd
        The password for the user that is being created.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    database
        The MongoDB database to use when checking if the user exists. Default is ``admin``.

    authdb
        The MongoDB database to use for authentication. Default is None.

    roles
        The roles that should be associated with the user. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.user_create <user_name> <user_password> <roles> <user> <password> <host> <port> <database>
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    if not roles:
        roles = []

    _roles = [{"role": _role, "db": database} for _role in roles]
    try:
        log.info("Creating user %s", name)
        mdb = pymongo.database.Database(conn, database)
        mdb.command("createUser", name, pwd=passwd, roles=_roles)

    except pymongo.errors.PyMongoError as err:
        log.error("Creating user %s failed with error: %s", name, err)
        return False
    return True


def user_remove(
    name, user=None, password=None, host=None, port=None, database="admin", authdb=None
):
    """
    Remove a MongoDB user

    name
        The name of the user that should be removed.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.user_remove <name> <user> <password> <host> <port> <database>
    """
    conn = _connect(user, password, host, port)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Removing user %s", name)
        mdb = pymongo.database.Database(conn, database)
        mdb.command("dropUser", name)
    except pymongo.errors.PyMongoError as err:
        log.error("Removing user %s failed with error: %s", name, err)
        return False
    return True


def user_roles_exists(
    name, roles, database, user=None, password=None, host=None, port=None, authdb=None
):
    """
    Checks if a user of a MongoDB database has specified roles

    name
        The name of the user to check for the specified roles.

    roles
        The roles to check are associated with the specified user.

    database
        The database to check has the specified roles for the specified user.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Examples:

    .. code-block:: bash

        salt '*' mongodb.user_roles_exists johndoe '["readWrite"]' dbname admin adminpwd localhost 27017

    .. code-block:: bash

        salt '*' mongodb.user_roles_exists johndoe '[{"role": "readWrite", "db": "dbname" }, {"role": "read", "db": "otherdb"}]' dbname admin adminpwd localhost 27017
    """
    try:
        roles = _to_dict(roles)
    except Exception:  # pylint: disable=broad-except
        return "Roles provided in wrong format"

    users = user_list(user, password, host, port, database, authdb)

    if isinstance(users, str):
        return "Failed to connect to mongo database"

    for user in users:
        if name == dict(user).get("user"):
            for role in roles:
                # if the role was provided in the shortened form, we convert it to a long form
                if not isinstance(role, dict):
                    role = {"role": role, "db": database}
                if role not in dict(user).get("roles", []):
                    return False
            return True

    return False


def user_grant_roles(
    name, roles, database, user=None, password=None, host=None, port=None, authdb=None
):
    """
    Grant one or many roles to a MongoDB user

    name
        The user to grant the specified roles to.

    roles
        The roles to grant to the specified user.

    database
        The database to great the roles against for the specified user.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Examples:

    .. code-block:: bash

        salt '*' mongodb.user_grant_roles johndoe '["readWrite"]' dbname admin adminpwd localhost 27017

    .. code-block:: bash

        salt '*' mongodb.user_grant_roles janedoe '[{"role": "readWrite", "db": "dbname" }, {"role": "read", "db": "otherdb"}]' dbname admin adminpwd localhost 27017
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        roles = _to_dict(roles)
    except Exception:  # pylint: disable=broad-except
        return "Roles provided in wrong format"

    try:
        log.info("Granting roles %s to user %s", roles, name)
        mdb = pymongo.database.Database(conn, database)
        mdb.command("grantRolesToUser", name, roles=roles)
    except pymongo.errors.PyMongoError as err:
        log.error(
            "Granting roles %s to user %s failed with error: %s", roles, name, err
        )
        return str(err)

    return True


def user_revoke_roles(
    name, roles, database, user=None, password=None, host=None, port=None, authdb=None
):
    """
    Revoke one or many roles to a MongoDB user

    user
        The user to connect to MongoDB as. Default is None.

    roles
        The roles to revoke from the specified user.

    database
        The database to revoke the roles from for the specified user.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Examples:

    .. code-block:: bash

        salt '*' mongodb.user_revoke_roles johndoe '["readWrite"]' dbname admin adminpwd localhost 27017

    .. code-block:: bash

        salt '*' mongodb.user_revoke_roles janedoe '[{"role": "readWrite", "db": "dbname" }, {"role": "read", "db": "otherdb"}]' dbname admin adminpwd localhost 27017
    """
    conn = _connect(user, password, host, port, authdb=authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        roles = _to_dict(roles)
    except Exception:  # pylint: disable=broad-except
        return "Roles provided in wrong format"

    try:
        log.info("Revoking roles %s from user %s", roles, name)
        mdb = pymongo.database.Database(conn, database)
        mdb.command("revokeRolesFromUser", name, roles=roles)
    except pymongo.errors.PyMongoError as err:
        log.error(
            "Revoking roles %s from user %s failed with error: %s", roles, name, err
        )
        return str(err)

    return True


def collection_create(
    collection,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    .. versionadded:: 3006.0

    Create a collection in the specified database.

    collection
        The name of the collection to create.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.collection_create mycollection <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Creating %s.%s", database, collection)
        mdb = pymongo.database.Database(conn, database)
        mdb.create_collection(collection)
    except pymongo.errors.PyMongoError as err:
        log.error(
            "Creating collection %r.%r failed with error %s", database, collection, err
        )
        return err
    return True


def collection_drop(
    collection,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    .. versionadded:: 3006.0

    Drop a collection in the specified database.

    collection
        The name of the collection to drop.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.collection_drop mycollection <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        log.info("Dropping %s.%s", database, collection)
        mdb = pymongo.database.Database(conn, database)
        mdb.drop_collection(collection)
    except pymongo.errors.PyMongoError as err:
        log.error(
            "Creating collection %r.%r failed with error %s", database, collection, err
        )
        return err
    return True


def collections_list(
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    .. versionadded:: 3006.0

    List the collections available in the specified database.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.collections_list mycollection <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        mdb = pymongo.database.Database(conn, database)
        ret = mdb.list_collection_names()
    except pymongo.errors.PyMongoError as err:
        log.error("Listing collections failed with error %s", err)
        return err
    return ret


def insert(
    objects,
    collection,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    Insert an object or list of objects into a collection

    objects
        The objects to insert into the collection, should be provided as a list.

    collection
        The collection to insert the objects into.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.insert '[{"foo": "FOO", "bar": "BAR"}, {"foo": "BAZ", "bar": "BAM"}]' mycollection <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        objects = _to_dict(objects)
    except Exception as err:  # pylint: disable=broad-except
        return err

    try:
        log.info("Inserting %r into %s.%s", objects, database, collection)
        mdb = pymongo.database.Database(conn, database)
        col = getattr(mdb, collection)
        ids = col.insert_many(objects)
        return ids.acknowledged
    except pymongo.errors.PyMongoError as err:
        log.error("Inserting objects %r failed with error %s", objects, err)
        return err


def update_one(
    objects,
    collection,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    Update an object into a collection
    http://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.update_one

    .. versionadded:: 2016.11.0

    objects
        The objects to update in the collection, should be provided as a list.

    collection
        The collection to insert the objects into.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.update_one '{"_id": "my_minion"} {"bar": "BAR"}' mycollection <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    objects = str(objects)
    objs = re.split(r"}\s+{", objects)

    if len(objs) != 2:
        return (
            "Your request does not contain a valid "
            + '\'{_"id": "my_id"} {"my_doc": "my_val"}\''
        )

    objs[0] = objs[0] + "}"
    objs[1] = "{" + objs[1]

    document = []

    for obj in objs:
        try:
            obj = _to_dict(obj)
            document.append(obj)
        except Exception as err:  # pylint: disable=broad-except
            return err

    _id_field = document[0]
    _update_doc = document[1]

    # need a string to perform the test, so using objs[0]
    test_f = find(collection, objs[0], user, password, host, port, database, authdb)
    if not isinstance(test_f, list):
        return "The find result is not well formatted. An error appears; cannot update."
    elif not test_f:
        return "Did not find any result. You should try an insert before."
    elif len(test_f) > 1:
        return "Too many results. Please try to be more specific."
    else:
        try:
            log.info("Updating %r into %s.%s", _id_field, database, collection)
            mdb = pymongo.database.Database(conn, database)
            col = getattr(mdb, collection)
            ids = col.update_one(_id_field, {"$set": _update_doc})
            nb_mod = ids.modified_count
            return f"{nb_mod} objects updated"
        except pymongo.errors.PyMongoError as err:
            log.error("Updating object %s failed with error %s", objects, err)
            return err


def find(
    collection,
    query=None,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    authdb=None,
):
    """
    Find an object or list of objects in a collection

    collection
        The collection to find the objects in.

    query
        The query to use when locating objects in the collection.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.find mycollection '[{"foo": "FOO", "bar": "BAR"}]' <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        query = _to_dict(query)
    except Exception as err:  # pylint: disable=broad-except
        return err

    try:
        log.info("Searching for %r in %s", query, collection)
        mdb = pymongo.database.Database(conn, database)
        col = getattr(mdb, collection)
        if isinstance(query, list):
            ret = []
            for _query in query:
                res = col.find(_query)
                _ret = [_res for _res in res]
                ret.extend(_ret)
        else:
            res = col.find(query)
            ret = [_res for _res in res]
        return ret
    except pymongo.errors.PyMongoError as err:
        log.error("Searching objects failed with error: %s", err)
        return err


def remove(
    collection,
    query=None,
    user=None,
    password=None,
    host=None,
    port=None,
    database="admin",
    w=1,
    authdb=None,
):
    """
    Remove an object or list of objects from a collection

    collection
        The collection to remove objects from based on the query.

    query
        Query to determine which objects to remove.

    user
        The user to connect to MongoDB as. Default is None.

    password
        The password to use to connect to MongoDB as.  Default is None.

    host
        The host where MongoDB is running. Default is None.

    port
        The host where MongoDB is running. Default is None.

    database
        The database where the collection is.

    w
        The number of matches to remove from the collection.

    authdb
        The MongoDB database to use for authentication. Default is None.

    CLI Example:

    .. code-block:: bash

        salt '*' mongodb.remove mycollection '[{"foo": "FOO", "bar": "BAR"}, {"foo": "BAZ", "bar": "BAM"}]' <user> <password> <host> <port> <database>

    """
    conn = _connect(user, password, host, port, database, authdb)
    if not conn:
        return "Failed to connect to mongo database"

    try:
        query = _to_dict(query)
    except Exception as err:  # pylint: disable=broad-except
        return _get_error_message(err)

    try:
        log.info("Removing %r from %s", query, collection)
        mdb = pymongo.database.Database(conn, database)
        col = getattr(mdb, collection)
        deleted_count = 0
        if isinstance(query, list):
            for _query in query:
                for count in range(0, w):
                    res = col.delete_one(_query)
                    deleted_count += res.deleted_count
        else:
            for count in range(0, w):
                res = col.delete_one(query)
                deleted_count += res.deleted_count
        return f"{deleted_count} objects removed"
    except pymongo.errors.PyMongoError as err:
        log.error("Removing objects failed with error: %s", _get_error_message(err))
        return _get_error_message(err)



"""
Util functions for the NXOS API modules.
"""

import json
import logging

import salt.utils.http
from salt.exceptions import SaltException
from salt.utils.args import clean_kwargs

log = logging.getLogger(__name__)

RPC_INIT_KWARGS = [
    "transport",
    "host",
    "username",
    "password",
    "port",
    "timeout",
    "verify",
    "rpc_version",
]


def _prepare_connection(**nxos_api_kwargs):
    """
    Prepare the connection with the remote network device, and clean up the key
    value pairs, removing the args used for the connection init.
    """
    nxos_api_kwargs = clean_kwargs(**nxos_api_kwargs)
    init_kwargs = {}
    # Clean up any arguments that are not required
    for karg, warg in nxos_api_kwargs.items():
        if karg in RPC_INIT_KWARGS:
            init_kwargs[karg] = warg
    if "host" not in init_kwargs:
        init_kwargs["host"] = "localhost"
    if "transport" not in init_kwargs:
        init_kwargs["transport"] = "https"
    if "port" not in init_kwargs:
        init_kwargs["port"] = 80 if init_kwargs["transport"] == "http" else 443
    verify = init_kwargs.get("verify", True)
    if isinstance(verify, bool):
        init_kwargs["verify_ssl"] = verify
    else:
        init_kwargs["ca_bundle"] = verify
    if "rpc_version" not in init_kwargs:
        init_kwargs["rpc_version"] = "2.0"
    if "timeout" not in init_kwargs:
        init_kwargs["timeout"] = 60
    return init_kwargs


def rpc(commands, method="cli", **kwargs):
    """
    Execute an arbitrary RPC request via the Nexus API.

    commands
        The commands to be executed.

    method: ``cli``
        The type of the response, i.e., raw text (``cli_ascii``) or structured
        document (``cli``). Defaults to ``cli`` (structured data).

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the connection device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.
    """
    init_args = _prepare_connection(**kwargs)
    log.error("These are the init args:")
    log.error(init_args)
    url = "{transport}://{host}:{port}/ins".format(
        transport=init_args["transport"], host=init_args["host"], port=init_args["port"]
    )
    headers = {"content-type": "application/json-rpc"}
    payload = []
    if not isinstance(commands, (list, tuple)):
        commands = [commands]
    for index, command in enumerate(commands):
        payload.append(
            {
                "jsonrpc": init_args["rpc_version"],
                "method": method,
                "params": {"cmd": command, "version": 1},
                "id": index + 1,
            }
        )
    opts = {"http_request_timeout": init_args["timeout"]}
    response = salt.utils.http.query(
        url,
        method="POST",
        opts=opts,
        data=json.dumps(payload),
        header_dict=headers,
        decode=True,
        decode_type="json",
        **init_args
    )
    if "error" in response:
        raise SaltException(response["error"])
    response_list = response["dict"]
    if isinstance(response_list, dict):
        response_list = [response_list]
    for index, command in enumerate(commands):
        response_list[index]["command"] = command
    return response_list



from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_INGRESS

parent = TC_H_INGRESS


def fix_msg(msg, kwarg):
    msg['handle'] = 0xFFFF0000


class options(nla):
    fields = (('value', 'I'),)



import inspect
import logging
import queue
import threading
import time
import traceback
from functools import partial

from pyroute2 import config

from . import schema
from .events import (
    DBMExitException,
    InvalidateHandlerException,
    RescheduleException,
    ShutdownException,
)
from .messages import cmsg, cmsg_event, cmsg_failed, cmsg_sstart

log = logging.getLogger(__name__)


def Events(*argv):
    for sequence in argv:
        if sequence is not None:
            for item in sequence:
                yield item


class NDBConfig(dict):
    def __init__(self, task_manager):
        self.task_manager = task_manager

    def __getitem__(self, key):
        return self.task_manager.config_get(key)

    def __setitem__(self, key, value):
        return self.task_manager.config_set(key, value)

    def __delitem__(self, key):
        return self.task_manager.config_del(key)

    def keys(self):
        return self.task_manager.config_keys()

    def items(self):
        return self.task_manager.config_items()

    def values(self):
        return self.task_manager.config_values()


class TaskManager:
    def __init__(self, ndb):
        self.ndb = ndb
        self.log = ndb.log
        self.event_map = {}
        self.event_queue = ndb._event_queue
        self.thread = None
        self.ctime = self.gctime = time.time()

    def register_handler(self, event, handler):
        if event not in self.event_map:
            self.event_map[event] = []
        self.event_map[event].append(handler)

    def unregister_handler(self, event, handler):
        self.event_map[event].remove(handler)

    def default_handler(self, target, event):
        if isinstance(getattr(event, 'payload', None), Exception):
            raise event.payload
        log.debug('unsupported event ignored: %s' % type(event))

    def check_sources_started(self, _locals, target, event):
        _locals['countdown'] -= 1
        if _locals['countdown'] == 0:
            self.ndb._dbm_ready.set()

    def wrap_method(self, method):
        #
        # this wrapper will be published in the DBM thread
        #
        def _do_local_generator(target, request):
            try:
                for item in method(*request.argv, **request.kwarg):
                    request.response.put(item)
                request.response.put(StopIteration())
            except Exception as e:
                request.response.put(e)

        def _do_local_single(target, request):
            try:
                (request.response.put(method(*request.argv, **request.kwarg)))
            except Exception as e:
                (request.response.put(e))

        #
        # this class will be used to map the requests
        #
        class cmsg_req(cmsg):
            def __init__(self, response, *argv, **kwarg):
                self['header'] = {'target': None}
                self.response = response
                self.argv = argv
                self.kwarg = kwarg

        #
        # this method will proxy the original one
        #
        def _do_dispatch_generator(self, *argv, **kwarg):
            if self.thread == id(threading.current_thread()):
                # same thread, run method locally
                for item in method(*argv, **kwarg):
                    yield item
            else:
                # another thread, run via message bus
                response = queue.Queue()
                request = cmsg_req(response, *argv, **kwarg)
                self.event_queue.put((request,))
                while True:
                    item = response.get()
                    if isinstance(item, StopIteration):
                        return
                    elif isinstance(item, Exception):
                        raise item
                    else:
                        yield item

        def _do_dispatch_single(self, *argv, **kwarg):
            if self.thread == id(threading.current_thread()):
                # same thread, run method locally
                return method(*argv, **kwarg)
            else:
                # another thread, run via message bus
                response = queue.Queue(maxsize=1)
                request = cmsg_req(response, *argv, **kwarg)
                self.event_queue.put((request,))
                ret = response.get()
                if isinstance(ret, Exception):
                    raise ret
                else:
                    return ret

        #
        # return the method spec to be announced
        #
        handler = _do_local_single
        proxy = _do_dispatch_single
        if inspect.isgeneratorfunction(method):
            handler = _do_local_generator
            proxy = _do_dispatch_generator
        return (cmsg_req, handler, proxy)

    def register_api(self, api_obj, prefix=''):
        for name in dir(api_obj):
            method = getattr(api_obj, name, None)
            if hasattr(method, 'publish'):
                if isinstance(method.publish, str):
                    name = method.publish
                name = f'{prefix}{name}'
                event, handler, proxy = self.wrap_method(method)
                setattr(self, name, partial(proxy, self))
                self.event_map[event] = [handler]

    def run(self):
        _locals = {'countdown': len(self.ndb._nl)}
        self.thread = id(threading.current_thread())

        # init the events map
        event_map = {
            cmsg_event: [lambda t, x: x.payload.set()],
            cmsg_failed: [lambda t, x: (self.ndb.schema.mark(t, 1))],
            cmsg_sstart: [partial(self.check_sources_started, _locals)],
        }
        self.event_map = event_map

        try:
            self.ndb.schema = schema.DBSchema(
                self.ndb.config,
                self.ndb.sources,
                self.event_map,
                self.log.channel('schema'),
            )
            self.register_api(self.ndb.schema, 'db_')
            self.register_api(self.ndb.schema.config, 'config_')
            self.ndb.bonfig = NDBConfig(self)

        except Exception as e:
            self.ndb._dbm_error = e
            self.ndb._dbm_ready.set()
            return

        for spec in self.ndb._nl:
            spec['event'] = None
            self.ndb.sources.add(**spec)

        for event, handlers in self.ndb.schema.event_map.items():
            for handler in handlers:
                self.register_handler(event, handler)

        stop = False
        source = None
        reschedule = []
        while not stop:
            source, events = self.event_queue.get()
            events = Events(events, reschedule)
            reschedule = []
            try:
                for event in events:
                    handlers = event_map.get(
                        event.__class__, [self.default_handler]
                    )

                    for handler in tuple(handlers):
                        try:
                            target = event['header']['target']
                            handler(target, event)
                        except RescheduleException:
                            if 'rcounter' not in event['header']:
                                event['header']['rcounter'] = 0
                            if event['header']['rcounter'] < 3:
                                event['header']['rcounter'] += 1
                                self.log.debug('reschedule %s' % (event,))
                                reschedule.append(event)
                            else:
                                self.log.error('drop %s' % (event,))
                        except InvalidateHandlerException:
                            try:
                                handlers.remove(handler)
                            except Exception:
                                self.log.error(
                                    'could not invalidate '
                                    'event handler:\n%s'
                                    % traceback.format_exc()
                                )
                        except ShutdownException:
                            stop = True
                            break
                        except DBMExitException:
                            return
                        except Exception:
                            self.log.error(
                                'could not load event:\n%s\n%s'
                                % (event, traceback.format_exc())
                            )
                    if time.time() - self.gctime > config.gc_timeout:
                        self.gctime = time.time()
            except Exception as e:
                self.log.error(f'exception <{e}> in source {source}')
                # restart the target
                try:
                    self.log.debug(f'requesting source {source} restart')
                    self.ndb.sources[source].state.set('restart')
                except KeyError:
                    self.log.debug(f'key error for {source}')
                    pass

        # release all the sources
        for target in tuple(self.ndb.sources.cache):
            source = self.ndb.sources.remove(target, sync=False)
            if source is not None and source.th is not None:
                self.log.debug(f'closing source {source}')
                source.close()
                if self.ndb.schema.config['db_cleanup']:
                    self.log.debug('flush DB for the target %s' % target)
                    self.ndb.schema.flush(target)
                else:
                    self.log.debug('leave DB for debug')

        # close the database
        self.ndb.schema.commit()
        self.ndb.schema.close()

        # close the logging
        for handler in self.log.logger.handlers:
            handler.close()



import inspect

try:
    from unittest import mock
except ImportError:
    mock = None

registry = []
use_mock = False


class LAB_API:
    def __init__(self, *argv, **kwarg):
        super().__init__(*argv, **kwarg)
        if use_mock:
            if mock is None:
                # postpone ImportError
                #
                # unittest may not be available on embedded platforms,
                # but it is still used by IPRoute class; it is safe
                # to leave it in the minimal for now, just raise an
                # exception when being used
                #
                # Bug-Url: https://github.com/svinota/pyroute2/pull/1096
                raise ImportError('unittest.mock not available')
            registry.append(self)
            for name, method in inspect.getmembers(
                self, predicate=inspect.ismethod
            ):
                setattr(self, name, mock.MagicMock(name=name, wraps=method))



"""
Module for sending messages via Telegram.

:configuration: In order to send a message via the Telegram, certain
    configuration is required in /etc/salt/minion on the relevant minions or
    in the pillar. Some sample configs might look like::

        telegram.chat_id: '123456789'
        telegram.token: '00000000:xxxxxxxxxxxxxxxxxxxxxxxx'

"""

import logging

from salt.exceptions import SaltInvocationError

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

log = logging.getLogger(__name__)

__virtualname__ = "telegram"


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    if not HAS_REQUESTS:
        return (False, "Missing dependency requests")
    return __virtualname__


def _get_chat_id():
    """
    Retrieves and return the Telegram's configured chat id

    :return:    String: the chat id string
    """
    chat_id = __salt__["config.get"]("telegram:chat_id") or __salt__["config.get"](
        "telegram.chat_id"
    )
    if not chat_id:
        raise SaltInvocationError("No Telegram chat id found")

    return chat_id


def _get_token():
    """
    Retrieves and return the Telegram's configured token

    :return:    String: the token string
    """
    token = __salt__["config.get"]("telegram:token") or __salt__["config.get"](
        "telegram.token"
    )
    if not token:
        raise SaltInvocationError("No Telegram token found")

    return token


def post_message(message, chat_id=None, token=None):
    """
    Send a message to a Telegram chat.

    :param message: The message to send to the Telegram chat.
    :param chat_id: (optional) The Telegram chat id.
    :param token:   (optional) The Telegram API token.
    :return:        Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' telegram.post_message message="Hello Telegram!"

    """
    if not chat_id:
        chat_id = _get_chat_id()

    if not token:
        token = _get_token()

    if not message:
        log.error("message is a required option.")

    return _post_message(message=message, chat_id=chat_id, token=token)


def _post_message(message, chat_id, token):
    """
    Send a message to a Telegram chat.

    :param chat_id:     The chat id.
    :param message:     The message to send to the telegram chat.
    :param token:       The Telegram API token.
    :return:            Boolean if message was sent successfully.
    """
    url = f"https://api.telegram.org/bot{token}/sendMessage"

    parameters = dict()
    if chat_id:
        parameters["chat_id"] = chat_id
    if message:
        parameters["text"] = message

    try:
        response = requests.post(url, data=parameters, timeout=120)
        result = response.json()

        log.debug("Raw response of the telegram request is %s", response)

    except Exception:  # pylint: disable=broad-except
        log.exception("Sending telegram api request failed")
        return False

    # Check if the Telegram Bot API returned successfully.
    if not result.get("ok", False):
        log.debug(
            "Sending telegram api request failed due to error %s (%s)",
            result.get("error_code"),
            result.get("description"),
        )
        return False

    return True



"""Global settings"""

import json
from pathlib import Path
from typing import Union
import yaml
from cpapis import uapi
from ._yaml import DumbYamlLoader

try:
    import paramiko as _

    HAS_PARAMIKO = True
except ImportError:
    HAS_PARAMIKO = False


def read_server_role() -> Union[tuple[None, None], tuple[str, str]]:
    """Read /etc/server.role and return imh_role, imh_class as a tuple of str"""
    try:
        server_role = Path('/etc/server.role').read_text(encoding='ascii')
    except OSError:
        imh_role, imh_class = None, None
    else:
        imh_role, imh_class = server_role.strip().split(':', maxsplit=1)
    return imh_role, imh_class


def get_secure_user(secure_fqdn: str) -> Union[str, None]:
    """Get the "secure" shared username from /etc/trueuserdomains"""
    try:
        user_domains: dict = yaml.load(
            Path('/etc/trueuserdomains').read_bytes(),
            DumbYamlLoader,
        )
    except OSError:  # If present, it's root.mail 640
        return None
    if not user_domains:
        return None
    return user_domains.get(secure_fqdn, None)


def rads_json():
    """Read rads.json"""
    # read users and groups from rads.json
    conf = json.loads(Path(__file__).parent.joinpath('rads.json').read_bytes())
    staff_groups: list[str] = conf['staff_groups']
    sys_users: list[str] = conf['sys_users']
    sys_mysql_users: list[str] = conf['sys_mysql_users']
    our_resellers: list[str] = conf['our_resellers']
    return staff_groups, sys_users, sys_mysql_users, our_resellers


def get_cpanel_extra_shared_ips() -> list[str]:
    """get any extra shared ips"""
    try:
        ipreasons = Path('/etc/reservedipreasons').read_text('ascii')
    except OSError:
        return []
    ips = []
    for line in ipreasons.splitlines():
        try:
            addr, reason = line.split('=', maxsplit=1)
        except ValueError:
            continue  # blank or malformed line
        if 'mail' in reason.lower():
            ips.append(addr.strip())
    return ips


def get_cpanel_inmotion_ips() -> list[str]:
    """get the shared inmotion ips"""
    try:
        return Path('/var/cpanel/mainips/inmotion').read_text('ascii').split()
    except OSError:
        return []


def get_secure_crt_info(secure_user: str, secure_fqdn: str) -> tuple[str, str]:
    """get the secure wildcard cert's ID and modulus"""
    if data := uapi(
        'SSL::installed_host',
        user=secure_user,
        args={'domain': secure_fqdn},
    )['result']['data']:
        return data['certificate']['id'], data['certificate']['modulus']
    raise AttributeError(f"No SSL certificates installed for {secure_user}")


def get_secure_key_file(secure_user: str, modulus: str) -> str:
    """get path to shared wildcard key"""
    if data := uapi('SSL::list_keys', secure_user)['result']['data']:
        for key_info in data:
            if key_info['modulus'] == modulus:
                key_id = key_info['id']
                return f'/home/{secure_user}/ssl/keys/{key_id}.key'
    raise AttributeError(f"No SSL keys installed for {secure_user}")



"""
Management of PostgreSQL tablespace
===================================

A module used to create and manage PostgreSQL tablespaces.

.. code-block:: yaml

    ssd-tablespace:
      postgres_tablespace.present:
        - name: indexes
        - directory: /mnt/ssd-data

.. versionadded:: 2015.8.0

"""

import salt.utils.dictupdate as dictupdate


def __virtual__():
    """
    Only load if the postgres module is present and is new enough (has ts funcs)
    """
    if "postgres.tablespace_exists" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    directory,
    options=None,
    owner=None,
    user=None,
    maintenance_db=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Ensure that the named tablespace is present with the specified properties.
    For more information about all of these options run ``man 7
    create_tablespace``.

    name
        The name of the tablespace to create/manage.

    directory
        The directory where the tablespace will be located, must already exist

    options
        A dictionary of options to specify for the tablespace.
        Currently, the only tablespace options supported are ``seq_page_cost``
        and ``random_page_cost``. Default values are shown in the example below:

        .. code-block:: yaml

            my_space:
              postgres_tablespace.present:
                - directory: /srv/my_tablespace
                - options:
                    seq_page_cost: 1.0
                    random_page_cost: 4.0

    owner
        The database user that will be the owner of the tablespace.
        Defaults to the user executing the command (i.e. the `user` option)

    user
        System user all operations should be performed on behalf of

    maintenance_db
        Database to act on

    db_user
        Database username if different from config or default

    db_password
        User password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Tablespace {name} is already present",
    }
    dbargs = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }
    tblspaces = __salt__["postgres.tablespace_list"](**dbargs)
    if name not in tblspaces:
        # not there, create it
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Tablespace {name} is set to be created"
            return ret
        if __salt__["postgres.tablespace_create"](
            name, directory, options, owner, **dbargs
        ):
            ret["comment"] = f"The tablespace {name} has been created"
            ret["changes"][name] = "Present"
            return ret

    # already exists, make sure it's got the right config
    if tblspaces[name]["Location"] != directory and not __opts__["test"]:
        ret["comment"] = (
            """Tablespace {} is not at the right location. This is
            unfixable without dropping and recreating the tablespace.""".format(
                name
            )
        )
        ret["result"] = False
        return ret

    if owner and not tblspaces[name]["Owner"] == owner:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Tablespace {name} owner to be altered"
        if (
            __salt__["postgres.tablespace_alter"](name, new_owner=owner)
            and not __opts__["test"]
        ):
            ret["comment"] = f"Tablespace {name} owner changed"
            ret["changes"][name] = {"owner": owner}
            ret["result"] = True

    if options:
        # options comes from postgres as a sort of json(ish) string, but it
        # can't really be parsed out, but it's in a fairly consistent format
        # that we should be able to string check:
        # {seq_page_cost=1.1,random_page_cost=3.9}
        # TODO remove options that exist if possible
        for k, v in options.items():
            # if 'seq_page_cost=1.1' not in '{seq_page_cost=1.1,...}'
            if f"{k}={v}" not in tblspaces[name]["Opts"]:
                if __opts__["test"]:
                    ret["result"] = None
                    ret["comment"] = (
                        """Tablespace {} options to be
                        altered""".format(
                            name
                        )
                    )
                    break  # we know it's going to be altered, no reason to cont
                if __salt__["postgres.tablespace_alter"](name, set_option={k: v}):
                    ret["comment"] = f"Tablespace {name} opts changed"
                    dictupdate.update(ret["changes"], {name: {"options": {k: v}}})
                    ret["result"] = True

    return ret


def absent(
    name,
    user=None,
    maintenance_db=None,
    db_user=None,
    db_password=None,
    db_host=None,
    db_port=None,
):
    """
    Ensure that the named tablespace is absent.

    name
        The name of the tablespace to remove

    user
        System user all operations should be performed on behalf of

    maintenance_db
        Database to act on

    db_user
        Database username if different from config or default

    db_password
        User password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    db_args = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }
    # check if tablespace exists and remove it
    if __salt__["postgres.tablespace_exists"](name, **db_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Tablespace {name} is set to be removed"
            return ret
        if __salt__["postgres.tablespace_remove"](name, **db_args):
            ret["comment"] = f"Tablespace {name} has been removed"
            ret["changes"][name] = "Absent"
            return ret

    # fallback
    ret["comment"] = "Tablespace {} is not present, so it cannot be removed".format(
        name
    )
    return ret



"""
The acl module handles publisher_acl operations

Additional information on publisher_acl can be
found by reading the salt documentation:

    https://docs.saltproject.io/en/latest/ref/publisheracl.html
"""

# Import python libraries

import salt.utils.stringutils


class PublisherACL:
    """
    Represents the publisher ACL and provides methods
    to query the ACL for given operations
    """

    def __init__(self, blacklist):
        self.blacklist = blacklist

    def user_is_blacklisted(self, user):
        """
        Takes a username as a string and returns a boolean. True indicates that
        the provided user has been blacklisted
        """
        return not salt.utils.stringutils.check_whitelist_blacklist(
            user, blacklist=self.blacklist.get("users", [])
        )

    def cmd_is_blacklisted(self, cmd):
        # If this is a regular command, it is a single function
        if isinstance(cmd, str):
            cmd = [cmd]
        for fun in cmd:
            if not salt.utils.stringutils.check_whitelist_blacklist(
                fun, blacklist=self.blacklist.get("modules", [])
            ):
                return True
        return False

    def user_is_whitelisted(self, user):
        return salt.utils.stringutils.check_whitelist_blacklist(
            user, whitelist=self.blacklist.get("users", [])
        )

    def cmd_is_whitelisted(self, cmd):
        # If this is a regular command, it is a single function
        if isinstance(cmd, str):
            cmd = [cmd]
        for fun in cmd:
            if salt.utils.stringutils.check_whitelist_blacklist(
                fun, whitelist=self.blacklist.get("modules", [])
            ):
                return True
        return False



# ===================================================================
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ===================================================================

"""Self-test suite for Cryptodome.Hash.cSHAKE128 and cSHAKE256"""

import unittest

from Cryptodome.SelfTest.loader import load_test_vectors
from Cryptodome.SelfTest.st_common import list_test_cases

from Cryptodome.Hash import cSHAKE128, cSHAKE256, SHAKE128, SHAKE256
from Cryptodome.Util.py3compat import b, bchr, tobytes


class cSHAKETest(unittest.TestCase):

    def test_left_encode(self):
        from Cryptodome.Hash.cSHAKE128 import _left_encode
        self.assertEqual(_left_encode(0), b'\x01\x00')
        self.assertEqual(_left_encode(1), b'\x01\x01')
        self.assertEqual(_left_encode(256), b'\x02\x01\x00')

    def test_bytepad(self):
        from Cryptodome.Hash.cSHAKE128 import _bytepad
        self.assertEqual(_bytepad(b'', 4), b'\x01\x04\x00\x00')
        self.assertEqual(_bytepad(b'A', 4), b'\x01\x04A\x00')
        self.assertEqual(_bytepad(b'AA', 4), b'\x01\x04AA')
        self.assertEqual(_bytepad(b'AAA', 4), b'\x01\x04AAA\x00\x00\x00')
        self.assertEqual(_bytepad(b'AAAA', 4), b'\x01\x04AAAA\x00\x00')
        self.assertEqual(_bytepad(b'AAAAA', 4), b'\x01\x04AAAAA\x00')
        self.assertEqual(_bytepad(b'AAAAAA', 4), b'\x01\x04AAAAAA')
        self.assertEqual(_bytepad(b'AAAAAAA', 4), b'\x01\x04AAAAAAA\x00\x00\x00')

    def test_new_positive(self):

        xof1 = self.cshake.new()
        xof2 = self.cshake.new(data=b("90"))
        xof3 = self.cshake.new().update(b("90"))

        self.assertNotEqual(xof1.read(10), xof2.read(10))
        xof3.read(10)
        self.assertEqual(xof2.read(10), xof3.read(10))

        xof1 = self.cshake.new()
        ref = xof1.read(10)
        xof2 = self.cshake.new(custom=b(""))
        xof3 = self.cshake.new(custom=b("foo"))

        self.assertEqual(ref, xof2.read(10))
        self.assertNotEqual(ref, xof3.read(10))

        xof1 = self.cshake.new(custom=b("foo"))
        xof2 = self.cshake.new(custom=b("foo"), data=b("90"))
        xof3 = self.cshake.new(custom=b("foo")).update(b("90"))

        self.assertNotEqual(xof1.read(10), xof2.read(10))
        xof3.read(10)
        self.assertEqual(xof2.read(10), xof3.read(10))

    def test_update(self):
        pieces = [bchr(10) * 200, bchr(20) * 300]
        h = self.cshake.new()
        h.update(pieces[0]).update(pieces[1])
        digest = h.read(10)
        h = self.cshake.new()
        h.update(pieces[0] + pieces[1])
        self.assertEqual(h.read(10), digest)

    def test_update_negative(self):
        h = self.cshake.new()
        self.assertRaises(TypeError, h.update, u"string")

    def test_digest(self):
        h = self.cshake.new()
        digest = h.read(90)

        # read returns a byte string of the right length
        self.assertTrue(isinstance(digest, type(b("digest"))))
        self.assertEqual(len(digest), 90)

    def test_update_after_read(self):
        mac = self.cshake.new()
        mac.update(b("rrrr"))
        mac.read(90)
        self.assertRaises(TypeError, mac.update, b("ttt"))

    def test_shake(self):
        # When no customization string is passed, results must match SHAKE
        for digest_len in range(64):
            xof1 = self.cshake.new(b'TEST')
            xof2 = self.shake.new(b'TEST')
            self.assertEqual(xof1.read(digest_len), xof2.read(digest_len))


class cSHAKE128Test(cSHAKETest):
    cshake = cSHAKE128
    shake = SHAKE128


class cSHAKE256Test(cSHAKETest):
    cshake = cSHAKE256
    shake = SHAKE256


class cSHAKEVectors(unittest.TestCase):
    pass


vector_files = [("ShortMsgSamples_cSHAKE128.txt", "Short Message Samples cSHAKE128", "128_cshake", cSHAKE128),
                ("ShortMsgSamples_cSHAKE256.txt", "Short Message Samples cSHAKE256", "256_cshake", cSHAKE256),
                ("CustomMsgSamples_cSHAKE128.txt", "Custom Message Samples cSHAKE128", "custom_128_cshake", cSHAKE128),
                ("CustomMsgSamples_cSHAKE256.txt", "Custom Message Samples cSHAKE256", "custom_256_cshake", cSHAKE256),
                ]

for file, descr, tag, test_class in vector_files:

    test_vectors = load_test_vectors(("Hash", "SHA3"), file, descr,
                                     {"len": lambda x: int(x),
                                      "nlen": lambda x: int(x),
                                      "slen": lambda x: int(x)}) or []

    for idx, tv in enumerate(test_vectors):
        if getattr(tv, "len", 0) == 0:
            data = b("")
        else:
            data = tobytes(tv.msg)
            assert(tv.len == len(tv.msg)*8)
        if getattr(tv, "nlen", 0) != 0:
            raise ValueError("Unsupported cSHAKE test vector")
        if getattr(tv, "slen", 0) == 0:
            custom = b("")
        else:
            custom = tobytes(tv.s)
            assert(tv.slen == len(tv.s)*8)

        def new_test(self, data=data, result=tv.md, custom=custom, test_class=test_class):
            hobj = test_class.new(data=data, custom=custom)
            digest = hobj.read(len(result))
            self.assertEqual(digest, result)

        setattr(cSHAKEVectors, "test_%s_%d" % (tag, idx), new_test)


def get_tests(config={}):
    tests = []
    tests += list_test_cases(cSHAKE128Test)
    tests += list_test_cases(cSHAKE256Test)
    tests += list_test_cases(cSHAKEVectors)
    return tests


if __name__ == '__main__':
    import unittest
    suite = lambda: unittest.TestSuite(get_tests())
    unittest.main(defaultTest='suite')



from Cryptodome.Util.number import long_to_bytes
from Cryptodome.PublicKey.ECC import EccKey


def _compute_ecdh(key_priv, key_pub):
    # See Section 5.7.1.2 in NIST SP 800-56Ar3
    pointP = key_pub.pointQ * key_priv.d
    if pointP.is_point_at_infinity():
        raise ValueError("Invalid ECDH point")
    z = long_to_bytes(pointP.x, pointP.size_in_bytes())
    return z


def key_agreement(**kwargs):
    """Perform a Diffie-Hellman key agreement.

    Keywords:
      kdf (callable):
        A key derivation function that accepts ``bytes`` as input and returns
        ``bytes``.
      static_priv (EccKey):
        The local static private key. Optional.
      static_pub (EccKey):
        The static public key that belongs to the peer. Optional.
      eph_priv (EccKey):
        The local ephemeral private key, generated for this session. Optional.
      eph_pub (EccKey):
        The ephemeral public key, received from the peer for this session. Optional.

    At least two keys must be passed, of which one is a private key and one
    a public key.

    Returns (bytes):
      The derived secret key material.
    """

    static_priv = kwargs.get('static_priv', None)
    static_pub = kwargs.get('static_pub', None)
    eph_priv = kwargs.get('eph_priv', None)
    eph_pub = kwargs.get('eph_pub', None)
    kdf = kwargs.get('kdf', None)

    if kdf is None:
        raise ValueError("'kdf' is mandatory")

    count_priv = 0
    count_pub = 0
    curve = None

    def check_curve(curve, key, name, private):
        if not isinstance(key, EccKey):
            raise TypeError("'%s' must be an ECC key" % name)
        if private and not key.has_private():
            raise TypeError("'%s' must be a private ECC key" % name)
        if curve is None:
            curve = key.curve
        elif curve != key.curve:
            raise TypeError("'%s' is defined on an incompatible curve" % name)
        return curve

    if static_priv is not None:
        curve = check_curve(curve, static_priv, 'static_priv', True)
        count_priv += 1

    if static_pub is not None:
        curve = check_curve(curve, static_pub, 'static_pub', False)
        count_pub += 1

    if eph_priv is not None:
        curve = check_curve(curve, eph_priv, 'eph_priv', True)
        count_priv += 1

    if eph_pub is not None:
        curve = check_curve(curve, eph_pub, 'eph_pub', False)
        count_pub += 1

    if (count_priv + count_pub) < 2 or count_priv == 0 or count_pub == 0:
        raise ValueError("Too few keys for the ECDH key agreement")

    Zs = b''
    Ze = b''

    if static_priv and static_pub:
        # C(*, 2s)
        Zs = _compute_ecdh(static_priv, static_pub)

    if eph_priv and eph_pub:
        # C(2e, 0s) or C(2e, 2s)
        if bool(static_priv) != bool(static_pub):
            raise ValueError("DH mode C(2e, 1s) is not supported")
        Ze = _compute_ecdh(eph_priv, eph_pub)
    elif eph_priv and static_pub:
        # C(1e, 2s) or C(1e, 1s)
        Ze = _compute_ecdh(eph_priv, static_pub)
    elif eph_pub and static_priv:
        # C(1e, 2s) or C(1e, 1s)
        Ze = _compute_ecdh(static_priv, eph_pub)

    Z = Ze + Zs

    return kdf(Z)



def ipdb_interfaces_view(ndb):
    '''Provide read-only interfaces view with IPDB layout.

    In addition to standard NDB fields provides some IPDB
    specific fields.

    The method returns a simple dict structure, no background
    updates or system changes are supported.

    Please open a ticket on the project page if you are
    missing any attribute used in your project:

    https://github.com/svinota/pyroute2/issues
    '''
    ret = {}

    for record in ndb.interfaces.dump():
        interface = record._as_dict()
        interface['ipdb_scope'] = 'system'
        interface['ipdb_priority'] = 0
        try:
            interface['ipaddr'] = tuple(
                (
                    (x.address, x.prefixlen)
                    for x in (
                        ndb.addresses.dump().select_records(index=record.index)
                    )
                )
            )
        except:
            with ndb.addresses.summary() as report:
                report.select_records(ifname=f"{record.ifname}")
                interface['ipaddr'] = tuple(
                    ((x.address, x.prefixlen) for x in report)
                )
        try:
            interface['ports'] = tuple(
                (
                    x.index
                    for x in (
                        ndb.interfaces.dump().select_records(
                            master=record.index
                        )
                    )
                )
            )
        except:
            with ndb.interfaces.dump() as report:
                report.select_records(ifname=f"{record.ifname}")
                interface['ports'] = tuple((x.index for x in report))
        try:
            interface['neighbours'] = tuple(
                (
                    x.dst
                    for x in (
                        ndb.neighbours.dump().select_records(
                            ifindex=record.index
                        )
                    )
                )
            )
        except:
            with ndb.neighbours.dump() as report:
                report.select_records(ifindex=record.index)
                interface['neighbours'] = tuple((x.dst for x in report))
        ret[record.ifname] = interface

    return ret



"""
An "Always Approved" eauth interface to test against, not intended for
production use
"""


def auth(username, password):  # pylint: disable=unused-argument
    """
    Authenticate!
    """
    return True



"""
Management of Zabbix user groups.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>

"""

__deprecated__ = (
    3009,
    "zabbix",
    "https://github.com/salt-extensions/saltext-zabbix",
)


def __virtual__():
    """
    Only make these states available if Zabbix module is available.
    """
    if "zabbix.usergroup_create" in __salt__:
        return True
    return (False, "zabbix module could not be loaded")


def present(name, **kwargs):
    """
    Creates new user group.
    NOTE: This function accepts all standard user group properties: keyword argument names differ depending on your
    zabbix version, see:
    https://www.zabbix.com/documentation/2.0/manual/appendix/api/usergroup/definitions#user_group

    .. versionadded:: 2016.3.0

    :param name: name of the user group
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        make_new_thai_monks_usergroup:
            zabbix_usergroup.present:
                - name: 'Thai monks'
                - gui_access: 1
                - debug_mode: 0
                - users_status: 0

    """
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_usergroup_created = f"User group {name} created."
    comment_usergroup_updated = f"User group {name} updated."
    comment_usergroup_notcreated = f"Unable to create user group: {name}. "
    comment_usergroup_exists = f"User group {name} already exists."
    changes_usergroup_created = {
        name: {
            "old": f"User group {name} does not exist.",
            "new": f"User group {name} created.",
        }
    }

    usergroup_exists = __salt__["zabbix.usergroup_exists"](name, **connection_args)

    if usergroup_exists:
        usergroup = __salt__["zabbix.usergroup_get"](name, **connection_args)[0]
        usrgrpid = int(usergroup["usrgrpid"])
        update_debug_mode = False
        update_gui_access = False
        update_users_status = False
        update_rights = False

        if "debug_mode" in kwargs:
            if int(kwargs["debug_mode"]) != int(usergroup["debug_mode"]):
                update_debug_mode = True

        if "gui_access" in kwargs:
            if int(kwargs["gui_access"]) != int(usergroup["gui_access"]):
                update_gui_access = True

        if "rights" in kwargs:
            # Older versions of Zabbix do not return the list of rights for the user group, handle this gracefully
            try:
                if usergroup["rights"]:
                    # Make sure right values are strings so we can compare them with the current user group rights
                    for right in kwargs["rights"]:
                        for key in right:
                            right[key] = str(right[key])
                    if sorted(kwargs["rights"]) != sorted(usergroup["rights"]):
                        update_rights = True
                else:
                    update_rights = True
            except KeyError:
                # As we don't know the current permissions, overwrite them as provided in the state.
                update_rights = True

        if "users_status" in kwargs:
            if int(kwargs["users_status"]) != int(usergroup["users_status"]):
                update_users_status = True

    # Dry run, test=true mode
    if __opts__["test"]:
        if usergroup_exists:
            if (
                update_debug_mode
                or update_gui_access
                or update_rights
                or update_users_status
            ):
                ret["result"] = None
                ret["comment"] = comment_usergroup_updated
            else:
                ret["result"] = True
                ret["comment"] = comment_usergroup_exists
        else:
            ret["result"] = None
            ret["comment"] = comment_usergroup_created
        return ret

    error = []

    if usergroup_exists:
        if (
            update_debug_mode
            or update_gui_access
            or update_rights
            or update_users_status
        ):
            ret["result"] = True
            ret["comment"] = comment_usergroup_updated

            if update_debug_mode:
                updated_debug = __salt__["zabbix.usergroup_update"](
                    usrgrpid, debug_mode=kwargs["debug_mode"], **connection_args
                )
                if "error" in updated_debug:
                    error.append(updated_debug["error"])
                else:
                    ret["changes"]["debug_mode"] = kwargs["debug_mode"]

            if update_gui_access:
                updated_gui = __salt__["zabbix.usergroup_update"](
                    usrgrpid, gui_access=kwargs["gui_access"], **connection_args
                )
                if "error" in updated_gui:
                    error.append(updated_gui["error"])
                else:
                    ret["changes"]["gui_access"] = kwargs["gui_access"]

            if update_rights:
                updated_rights = __salt__["zabbix.usergroup_update"](
                    usrgrpid, rights=kwargs["rights"], **connection_args
                )
                if "error" in updated_rights:
                    error.append(updated_rights["error"])
                else:
                    ret["changes"]["rights"] = kwargs["rights"]

            if update_users_status:
                updated_status = __salt__["zabbix.usergroup_update"](
                    usrgrpid, users_status=kwargs["users_status"], **connection_args
                )
                if "error" in updated_status:
                    error.append(updated_status["error"])
                else:
                    ret["changes"]["users_status"] = kwargs["users_status"]

        else:
            ret["result"] = True
            ret["comment"] = comment_usergroup_exists
    else:
        usergroup_create = __salt__["zabbix.usergroup_create"](name, **kwargs)

        if "error" not in usergroup_create:
            ret["result"] = True
            ret["comment"] = comment_usergroup_created
            ret["changes"] = changes_usergroup_created
        else:
            ret["result"] = False
            ret["comment"] = comment_usergroup_notcreated + str(
                usergroup_create["error"]
            )

    # error detected
    if error:
        ret["changes"] = {}
        ret["result"] = False
        ret["comment"] = str(error)

    return ret


def absent(name, **kwargs):
    """
    Ensures that the user group does not exist, eventually delete user group.

    .. versionadded:: 2016.3.0

    :param name: name of the user group
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        delete_thai_monks_usrgrp:
            zabbix_usergroup.absent:
                - name: 'Thai monks'
    """
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_usergroup_deleted = f"User group {name} deleted."
    comment_usergroup_notdeleted = f"Unable to delete user group: {name}. "
    comment_usergroup_notexists = f"User group {name} does not exist."
    changes_usergroup_deleted = {
        name: {
            "old": f"User group {name} exists.",
            "new": f"User group {name} deleted.",
        }
    }

    usergroup_exists = __salt__["zabbix.usergroup_exists"](name, **connection_args)

    # Dry run, test=true mode
    if __opts__["test"]:
        if not usergroup_exists:
            ret["result"] = True
            ret["comment"] = comment_usergroup_notexists
        else:
            ret["result"] = None
            ret["comment"] = comment_usergroup_deleted
        return ret

    usergroup_get = __salt__["zabbix.usergroup_get"](name, **connection_args)

    if not usergroup_get:
        ret["result"] = True
        ret["comment"] = comment_usergroup_notexists
    else:
        try:
            usrgrpid = usergroup_get[0]["usrgrpid"]
            usergroup_delete = __salt__["zabbix.usergroup_delete"](
                usrgrpid, **connection_args
            )
        except KeyError:
            usergroup_delete = False

        if usergroup_delete and "error" not in usergroup_delete:
            ret["result"] = True
            ret["comment"] = comment_usergroup_deleted
            ret["changes"] = changes_usergroup_deleted
        else:
            ret["result"] = False
            ret["comment"] = comment_usergroup_notdeleted + str(
                usergroup_delete["error"]
            )

    return ret



"""
Scan a netmask or ipaddr for open ssh ports
"""

import copy
import logging
import socket

import salt.utils.network
from salt._compat import ipaddress

log = logging.getLogger(__name__)


def targets(tgt, tgt_type="glob", **kwargs):
    """
    Return the targets from the flat yaml file, checks opts for location but
    defaults to /etc/salt/roster
    """
    rmatcher = RosterMatcher(tgt, tgt_type)
    return rmatcher.targets()


class RosterMatcher:
    """
    Matcher for the roster data structure
    """

    def __init__(self, tgt, tgt_type):
        self.tgt = tgt
        self.tgt_type = tgt_type

    def targets(self):
        """
        Return ip addrs based on netmask, sitting in the "glob" spot because
        it is the default
        """
        addrs = []
        ret = {}
        ports = __opts__["ssh_scan_ports"]
        if not isinstance(ports, list):
            # Comma-separate list of integers
            ports = list(map(int, str(ports).split(",")))
        if self.tgt_type == "list":
            tgts = self.tgt
        else:
            tgts = [self.tgt]
        for tgt in tgts:
            try:
                addrs.append(ipaddress.ip_address(tgt))
            except ValueError:
                try:
                    addrs.extend(ipaddress.ip_network(tgt).hosts())
                except ValueError:
                    pass
        for addr in addrs:
            addr = str(addr)
            ret[addr] = copy.deepcopy(__opts__.get("roster_defaults", {}))
            log.trace("Scanning host: %s", addr)
            for port in ports:
                log.trace("Scanning port: %s", port)
                try:
                    sock = salt.utils.network.get_socket(addr, socket.SOCK_STREAM)
                    sock.settimeout(float(__opts__["ssh_scan_timeout"]))
                    sock.connect((addr, port))
                    sock.shutdown(socket.SHUT_RDWR)
                    sock.close()
                    ret[addr].update({"host": addr, "port": port})
                except OSError:
                    pass
        return ret



"""
This is the default range matcher.
"""

import logging

HAS_RANGE = False
try:
    import seco.range

    HAS_RANGE = True
except ImportError:
    pass

log = logging.getLogger(__name__)


def match(tgt, opts=None, minion_id=None):
    """
    Matches based on range cluster
    """
    if not opts:
        opts = __opts__
    if HAS_RANGE:
        range_ = seco.range.Range(opts["range_server"])
        try:
            return opts["grains"]["fqdn"] in range_.expand(tgt)
        except seco.range.RangeException as exc:
            log.debug("Range exception in compound match: %s", exc)
            return False
    return False



"""Functions for acquiring a process lock"""
import socket
import hashlib
import contextlib
import time
from typing import Optional
import __main__ as main_obj


class LockError(Exception):
    """Raised when a rads.lock cannot be acquired"""

    __module__ = 'rads'


@contextlib.contextmanager
def lock(name: Optional[str] = None):
    """Context manager to create an abstract UNIX socket which will behave as
    a "lock". This can be used en lieu of /var/run file pids/locks to determine
    if an instance of a process is already running and ensure the lock is not
    left behind if the process terminates prematurely

    Args:
        name: unique name of the lock; defaults to filename of ``__main__``
    Raises:
        LockError: if the "lock" already existed
    """
    if name is None:
        # caveat: this will raise AttributeError
        # if run in an interactive python shell
        name = main_obj.__file__
    name = name.encode('utf-8')
    lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        lock_socket.bind('\0%s' % hashlib.sha256(name).hexdigest())
    except socket.error as exc:
        raise LockError(name) from exc
    try:
        yield
    finally:
        lock_socket.close()


lock.__module__ = 'rads'


@contextlib.contextmanager
def wait_lock(
    name: Optional[str] = None,
    sleep: float = 0.2,
    max_sleep: Optional[int] = None,
):
    """Context manager to wait until a rads.lock() can be acquired, and do so

    Args:
        name: unique name of the lock; defaults to filename of ``__main__``
        sleep: secs before trying to acquire the lock again
        max_sleep: max number of tries to acquire the lock
    Raises:
        LockError: if max_sleep is exceeded
    """
    slept = 0.0
    while True:
        try:
            with lock(name):
                yield
                return
        except LockError:
            time.sleep(sleep)
            slept += sleep
            if max_sleep is not None and slept >= max_sleep:
                raise


wait_lock.__module__ = 'rads'



"""
    salt._logging
    ~~~~~~~~~~~~~

    This is salt's new logging setup.
    As the name suggests, this is considered an internal API which can change without notice,
    although, as best effort, we'll try not to break code relying on it.

    The ``salt._logging`` package should be imported as soon as possible since salt tweaks
    the python's logging system.
"""

from salt._logging.impl import (
    DFLT_LOG_DATEFMT,
    DFLT_LOG_DATEFMT_LOGFILE,
    DFLT_LOG_FMT_CONSOLE,
    DFLT_LOG_FMT_JID,
    DFLT_LOG_FMT_LOGFILE,
    LOG_COLORS,
    LOG_LEVELS,
    LOG_VALUES_TO_LEVELS,
    SORTED_LEVEL_NAMES,
    freeze_logging_options_dict,
    get_console_handler,
    get_extended_logging_handlers,
    get_log_record_factory,
    get_logfile_handler,
    get_logging_level_from_string,
    get_logging_options_dict,
    get_lowest_log_level,
    get_temp_handler,
    in_mainprocess,
    is_console_handler_configured,
    is_extended_logging_configured,
    is_logfile_handler_configured,
    is_temp_handler_configured,
    set_log_record_factory,
    set_logging_options_dict,
    set_lowest_log_level,
    set_lowest_log_level_by_opts,
    setup_console_handler,
    setup_extended_logging,
    setup_log_granular_levels,
    setup_logfile_handler,
    setup_logging,
    setup_temp_handler,
    shutdown_console_handler,
    shutdown_extended_logging,
    shutdown_logfile_handler,
    shutdown_logging,
    shutdown_temp_handler,
)
